/*
	Trig Wars.
	Copyright (c) 2008, Darragh Coy.
	All rights reserved.
	Original work(s), source code & author(s) details available at: http://www.darraghcoy.com

	Redistribution and use in source and binary forms, with or without modification, are permitted 
	provided that the following conditions are met:

		* 	Redistributions of source code must retain the above copyright notice, this list of 
			conditions and the following disclaimer.
		* 	Redistributions in binary form must reproduce the above copyright notice, this list of 
			conditions and the following disclaimer in the documentation.
		* 	Neither the name of the author nor the names of contributors may be used to endorse or 
			promote products derived from this software without specific prior written permission.
		* 	The use of any third party libraries, modules or data files provided with the source 
			form of the program must follow and respect the individual licensing terms of these 
			components. You do not have permission to modify or circumvent the licensing terms for 
			provided third party components unless the third party licenses specifically allow this. 
			If you do not agree to the terms of provided third party components then you do not have 
			permission to use them in your derived work(s).
		* 	The terms of this license do not apply to separate (but related) distributions if supplied 
			by the author(s), only to the contents of this source distribution, derived works, and the 
			binaries produced by either modified or unmodified versions of the source code. The terms 
			of this license do not also apply to any third party libraries or components provided with 
			the source distribution. Such third party elements are licensed under their own separate 
			agreements.

	THIS SOFTWARE IS PROVIDED BY DARRAGH COY ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
	BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL DARRAGH COY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
	IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef HIGH_RESOLUTION_TIMER_H
#define HIGH_RESOLUTION_TIMER_H

//=====================================================================================================================
//=====================================================================================================================

// Trim windows fat

#ifndef WIN32_LEAN_AND_MEAN

	#define WIN32_LEAN_AND_MEAN

#endif

#include "Base\Types.h"
#include <Windows.h>

//=====================================================================================================================
//=====================================================================================================================

namespace Game
{
	//=================================================================================================================
	/// @brief		This static class creates a high resolution timing system. Time values are read off one core exclusively
	///				in a created thread because of the inconsitency of timing values across cores in mult-core systems.
	///				This is used by the application core to time when frame updates should be applied.
	//=================================================================================================================

	class HighResolutionTimer
	{
		public:
			
			//=========================================================================================================
			// Functions
			//=========================================================================================================
		
				/// Initializes the high resolution timing system: must be called before high resolution timing can be used

				static void Initialize();

				/// Shuts down the high resolution timing system

				static void Shutdown();
				
				/// Updates the current time stamp and saves the previous stamp, thus allowing the time delta between calls to this function to be obtained
				
				static void Update();
				
				/// Updates both the current and previous time stamps with the current time stamp value, thus invalidating any time passed
				
				static void Reset();
				
				/// Obtains the time in seconds between the last two calls to Update(), or returns 0 if Reset() was called last:
				
				static float GetTimeDelta();

		private:
		
			//=========================================================================================================
			// Functions
			//=========================================================================================================
			
				/// Main loop for the timing thread: continuosly updates the timing values
			
				static DWORD WINAPI TimingThreadMain( void * user_data );

			//=========================================================================================================
			// Variables
			//=========================================================================================================
			
				/// The current time stamp:
				
				static LARGE_INTEGER s_current_time_stamp;
				
				/// The previous time stamp:
				
				static LARGE_INTEGER s_previous_time_stamp;
				
				/// The frequency of the high resolution timer:
				
				static LARGE_INTEGER s_timer_frequency;
				
				/// The timing thread
				
				static HANDLE s_timing_thread;
				
				/// The timing thread id
				
				static DWORD s_timing_thread_id;
				
				/// Mutex for timing values:
				
				static HANDLE s_mutex;
				
				/// Continue running the timing thread ?
				
				static bool s_run_timing_thread;
				
				/// The current time delta between calls to Update()
				
				static float s_current_time_delta;
	};
};

#endif
