/*
	Trig Wars.
	Copyright (c) 2008, Darragh Coy.
	All rights reserved.
	Original work(s), source code & author(s) details available at: http://www.darraghcoy.com

	Redistribution and use in source and binary forms, with or without modification, are permitted 
	provided that the following conditions are met:

		* 	Redistributions of source code must retain the above copyright notice, this list of 
			conditions and the following disclaimer.
		* 	Redistributions in binary form must reproduce the above copyright notice, this list of 
			conditions and the following disclaimer in the documentation.
		* 	Neither the name of the author nor the names of contributors may be used to endorse or 
			promote products derived from this software without specific prior written permission.
		* 	The use of any third party libraries, modules or data files provided with the source 
			form of the program must follow and respect the individual licensing terms of these 
			components. You do not have permission to modify or circumvent the licensing terms for 
			provided third party components unless the third party licenses specifically allow this. 
			If you do not agree to the terms of provided third party components then you do not have 
			permission to use them in your derived work(s).
		* 	The terms of this license do not apply to separate (but related) distributions if supplied 
			by the author(s), only to the contents of this source distribution, derived works, and the 
			binaries produced by either modified or unmodified versions of the source code. The terms 
			of this license do not also apply to any third party libraries or components provided with 
			the source distribution. Such third party elements are licensed under their own separate 
			agreements.

	THIS SOFTWARE IS PROVIDED BY DARRAGH COY ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
	BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL DARRAGH COY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
	IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef GAME_COLOR_H
#define GAME_COLOR_H

//=====================================================================================================================
//=====================================================================================================================

#include "Base\Math\Vector2.h"
#include "Base\Math\Vector3.h"
#include "Base\Math\Vector4.h"
#include "Base\Assertion.h"
#include <cmath>

// Specific:

#ifdef GAME_SUPPORT_DIRECT_X_TYPES

	#include <D3DX10Math.h>

#endif

//=====================================================================================================================
//=====================================================================================================================

namespace Game
{
	//=================================================================================================================
	/// @brief		Represents a four component floating point color.
	//=================================================================================================================

	struct Color
	{	
		public:

			//=========================================================================================================
			// Functions
			//=========================================================================================================

				//=====================================================================================================
				/// @brief		Constructor. Creates the color.
				/// @param		r - red component of the color or red
				/// @param		g - green component of the color or green
				/// @param		b - blue component of the color or blue
				/// @param		a - alpha component of the color or alpha
				//=====================================================================================================
				
				inline Color( float r = 1.0f , float g = 1.0f , float b = 1.0f , float a = 1.0f )
				{
					this->r = r;
					this->g = g;
					this->b = b;
					this->a = a;
				}						
				
				//=====================================================================================================
				/// @brief		Constructor. Creates the color from one 32-bit integer. 
				/// @param		c - 32-bit integer representing the color
				/// @remarks	The hex RGBA components of the integer are laid out as follows:
				///				0xRRGGBBAA. So the most significant byte is red and the least is alpha.
				//=====================================================================================================
				
				inline Color( unsigned c )
				{								
					const float DIV = 1.0f / 255.0f;
					
					// Retrive rgba components:
					
					unsigned r = ( c >> 24 ) & 0xff;
					unsigned g = ( c >> 16 ) & 0xff;
					unsigned b = ( c >> 8  ) & 0xff;
					
					unsigned a = c & 0xff;
				
					this->r = float(r) * DIV;
					this->g = float(g) * DIV;
					this->b = float(b) * DIV;
					this->a = float(a) * DIV;					
				}				
	
				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Constructor. Creates the vector from a given directx 2d vector and sets blue to 0, 
				///				and alpha to 1
				/// @param		v - vector to create from
				//=====================================================================================================				
	
				inline Color( const D3DXVECTOR2 & v )
				{
					this->r = v.x;
					this->g = v.y;
					this->b = 0;
					this->a = 1;
				}					

				#endif

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Constructor. Creates the vector from a given directx 3d vector and sets alpha to 1.
				/// @param		v - vector to create from
				//=====================================================================================================				
	
				inline Color( const D3DXVECTOR3 & v )
				{
					this->r = v.x;
					this->g = v.y;
					this->b = v.z;
					this->a = 1.0f;
				}					

				#endif

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Constructor. Creates the color from a given directx 4d vector
				/// @param		v - vector to create from
				//=====================================================================================================				
	
				inline Color( const D3DXVECTOR4 & v )
				{
					this->r = v.x;
					this->g = v.y;
					this->b = v.z;
					this->a = v.w;
				}
				
				#endif
				
				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Constructor. Creates the color from a given directx color
				/// @param		c - color to create from
				//=====================================================================================================				
	
				inline Color( const D3DXCOLOR & c )
				{
					this->r = c.r;
					this->g = c.g;
					this->b = c.b;
					this->a = c.a;
				}				

				#endif
				
				//=====================================================================================================
				/// @brief		Copy constructor. Creates this color from another color.
				/// @param		c - other color to construct from
				//=====================================================================================================
				
				inline Color( const Color & c )
				{
					this->r = c.r;
					this->g = c.g;
					this->b = c.b;
					this->a = c.a;
				}					
				
				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the other vector to this vector.
				/// @param		v - other vector to assign from
				/// @returns	This color
				//=====================================================================================================
				
				template<class VT> inline Color & operator = ( const Vector4<VT> & v )
				{
					this->r = (float) v.x;
					this->g = (float) v.y;
					this->b = (float) v.z;
					this->a = (float) v.w;
					
					return * this;
				}
				
				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the 3d vector to this vector and sets alpha to 1
				/// @param		v - other vector to assign from
				/// @returns	This color
				//=====================================================================================================
				
				template<class VT> inline Color & operator = ( const Vector3<VT> & v )
				{
					this->r = (float) v.x;
					this->g = (float) v.y;
					this->b = (float) v.z;
					this->a = 1.0f;
					
					return * this;
				}
				
				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the 2d vector to this vector and sets blue to 1, 
				///				and alpha to 1
				/// @param		v - other vector to assign from
				/// @returns	This color
				//=====================================================================================================
				
				template<class VT> inline Color & operator = ( const Vector2<VT> & v )
				{
					this->r = (float) v.x;
					this->g = (float) v.y;
					this->b = 1.0f;
					this->a = 1.0f;
					
					return * this;
				}											
				
				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the given four component color array to this vector,
				/// @param		a - array to assign to this vector.
				/// @returns	This color
				//=====================================================================================================
				
				inline Color & operator = ( const float * a )
				{				
					this->r = a[0];
					this->g = a[1];
					this->b = a[2];
					this->a = a[3];
					
					return * this;
				}
				
				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the color from one 32-bit integer. 
				/// @param		c - 32-bit integer representing the color
				/// @returns	This color
				/// @remarks	The hex RGBA components of the integer are laid out as follows:
				///				0xRRGGBBAA. So the most significant byte is red and the least is alpha.
				//=====================================================================================================
				
				inline Color & operator = ( unsigned c )
				{								
					const float DIV = 1.0f / 255.0f;
					
					// Retrive rgba components:
					
					unsigned r = ( c >> 24 ) & 0xff;
					unsigned g = ( c >> 16 ) & 0xff;
					unsigned b = ( c >> 8  ) & 0xff;
					
					unsigned a = c & 0xff;
				
					this->r = float(r) * DIV;
					this->g = float(g) * DIV;
					this->b = float(b) * DIV;
					this->a = float(a) * DIV;					
				}				

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the given 2d directx vector to this vector and sets blue
				///				to 1 and alpha to 1.
				/// @param		v - vector to assign
				/// @returns	This color
				//=====================================================================================================				
	
				inline Color & operator = ( const D3DXVECTOR2 & v )
				{
					this->r = (float) v.x;
					this->g = (float) v.y;
					this->b = 1.0f;
					this->a = 1.0f;

					return * this;
				}					

				#endif

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the given 3d directx vector to this vector, and sets alpha
				///				to 1.
				/// @param		v - vector to assign
				/// @returns	This color
				//=====================================================================================================				
	
				inline Color & operator = ( const D3DXVECTOR3 & v )
				{
					this->r = (float) v.x;
					this->g = (float) v.y;
					this->b = (float) v.z;
					this->a = 1.0f;

					return * this;
				}					

				#endif

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the given 4d directx vector to this vector.
				/// @param		v - vector to assign
				/// @returns	This color
				//=====================================================================================================				
	
				inline Color & operator = ( const D3DXVECTOR4 & v )
				{
					this->r = (float) v.x;
					this->g = (float) v.y;
					this->b = (float) v.z;
					this->a = (float) v.w;

					return * this;
				}
				
				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the given directx color to this color.
				/// @param		c - color to assign
				/// @returns	This color
				//=====================================================================================================				
	
				inline Color & operator = ( const D3DXCOLOR & c )
				{
					this->r = (float) c.r;
					this->g = (float) c.g;
					this->b = (float) c.b;
					this->a = (float) c.a;

					return * this;
				}								

				#endif
				
				//=====================================================================================================
				/// @brief		Array index accessor. Returns a reference of the color at the given index.
				/// @param		i - color element to access
				/// @returns	chosen color element
				/// @pre		Given index must be in range
				/// @remarks	i = 0 is red, i = 1 is blue, i = 2 is green , i = 3 is alpha
				//=====================================================================================================
				
				inline float operator[]( unsigned i ) const
				{
					ENSURE(i<4,"Given index must be in range");
				
					return *((&r)+i);
				}
				
				//=====================================================================================================
				/// @brief		Array index accessor. Returns a reference of the color at the given index.
				/// @param		i - color element to access
				/// @returns	chosen color element, ready for writing. 
				/// @pre		Given index must be in range
				/// @remarks	i = 0 is red, i = 1 is blue, i = 2 is green , i = 3 is alpha
				//=====================================================================================================
				
				inline float & operator[]( unsigned i )
				{
					ENSURE(i<4,"Given index must be in range");
				
					return *((&r)+i);
				}																																											
				
				//=====================================================================================================
				/// @brief		Scalar addition. Adds this color's RGB values by the given amount and returns a new
				///				vector with the result.
				/// @param		value - value to add by
				/// @returns	New color with the result
				/// @remarks	The alpha value of the new vector remains the same as this vector
				//=====================================================================================================
				
				template <class ST> inline Color AddedRGB( ST value ) const
				{
					// Return the answer:
					
					return Color
					(
						(float)( this->r + value ),
						(float)( this->g + value ),
						(float)( this->b + value ),
						this->a
					);
				}
				
				//=====================================================================================================
				/// @brief		Scalar addition. Adds this color's RGBA values by the given amount and returns a new
				///				vector with the result.
				/// @param		value - value to add by
				/// @returns	New color with the result
				//=====================================================================================================
				
				template <class ST> inline Color AddedRGBA( ST value ) const
				{
					// Return the answer:
					
					return Color
					(
						(float)( this->r + value ),
						(float)( this->g + value ),
						(float)( this->b + value ),
						(float)( this->a + value )
					);
				}				
				
				//=====================================================================================================
				/// @brief		Scalar subtraction. Subtracts this color's RGB values by the given amount and returns 
				///				a new vector with the result.
				/// @param		value - value to subtract by
				/// @returns	New color with the result
				/// @remarks	The alpha value of the new vector remains the same as this vector
				//=====================================================================================================
				
				template <class ST> inline Color SubedRGB( ST value ) const
				{
					// Return the answer:
					
					return Color
					(
						(float)( this->r - value ),
						(float)( this->g - value ),
						(float)( this->b - value ),
						this->a
					);
				}
				
				//=====================================================================================================
				/// @brief		Scalar subtraction. Subtracts this color's RGBA values by the given amount and returns 
				///				a new vector with the result.
				/// @param		value - value to subtract by
				/// @returns	New color with the result
				//=====================================================================================================
				
				template <class ST> inline Color SubedRGBA( ST value ) const
				{
					// Return the answer:
					
					return Color
					(
						(float)( this->r - value ),
						(float)( this->g - value ),
						(float)( this->b - value ),
						this->a
					);
				}
				
				//=====================================================================================================
				/// @brief		Scalar division. Subtracts this color's RGB scales by the given amount and returns 
				///				a new vector with the result.
				/// @param		scale - scale to divide by
				/// @returns	New color with the result
				/// @remarks	The alpha scale of the new vector remains the same as this vector
				/// @pre		Scale must not be zero
				//=====================================================================================================
				
				template <class ST> inline Color DivedRGB( ST scale ) const
				{
					ENSURE( scale != 0 , "Scale must not be zero" );
				
					// Return the answer:
					
					return Color
					(
						(float)( this->r / scale ),
						(float)( this->g / scale ),
						(float)( this->b / scale ),
						this->a
					);
				}
				
				//=====================================================================================================
				/// @brief		Scalar division. Subtracts this color's RGBA scales by the given amount and returns 
				///				a new vector with the result.
				/// @param		scale - scale to divide by
				/// @returns	New color with the result
				/// @pre		Scale must not be zero
				//=====================================================================================================
				
				template <class ST> inline Color DivedRGBA( ST scale ) const
				{
					ENSURE( scale != 0 , "Scale must not be zero" );
				
					// Return the answer:
					
					return Color
					(
						(float)( this->r / scale ),
						(float)( this->g / scale ),
						(float)( this->b / scale ),
						this->a
					);
				}
				
				//=====================================================================================================
				/// @brief		Scalar multiplication. Subtracts this color's RGB scales by the given amount and returns 
				///				a new vector with the result.
				/// @param		scale - scale to multiply by
				/// @returns	New color with the result
				/// @remarks	The alpha scale of the new vector remains the same as this vector
				//=====================================================================================================
				
				template <class ST> inline Color MuledRGB( ST scale ) const
				{
					// Return the answer:
					
					return Color
					(
						(float)( this->r * scale ),
						(float)( this->g * scale ),
						(float)( this->b * scale ),
						this->a
					);
				}
				
				//=====================================================================================================
				/// @brief		Scalar multiplication. Subtracts this color's RGBA scales by the given amount and returns 
				///				a new vector with the result.
				/// @param		scale - scale to multiply by
				/// @returns	New color with the result
				//=====================================================================================================
				
				template <class ST> inline Color MuledRGBA( ST scale ) const
				{
					// Return the answer:
					
					return Color
					(
						(float)( this->r * scale ),
						(float)( this->g * scale ),
						(float)( this->b * scale ),
						this->a
					);
				}																			
				
				//=====================================================================================================
				/// @brief		Scalar addition. Adds this color's RGB values by the given amount.
				/// @param		value - value to add by
				//=====================================================================================================
				
				template <class ST> inline void AddRGB( ST value )
				{
					this->r = (float)( this->r + value );
					this->g = (float)( this->g + value );
					this->b = (float)( this->b + value );
				}						
				
				//=====================================================================================================
				/// @brief		Scalar addition. Adds this color's RGBA values by the given amount.
				/// @param		value - value to add by
				//=====================================================================================================
				
				template <class ST> inline void AddRGBA( ST value )
				{
					this->r = (float)( this->r + value );
					this->g = (float)( this->g + value );
					this->b = (float)( this->b + value );
					this->a = (float)( this->a + value );
				}				
				
				//=====================================================================================================
				/// @brief		Scalar subtraction. Subtracts this color's RGB values by the given amount.
				/// @param		value - value to subtract by
				//=====================================================================================================
				
				template <class ST> inline void SubRGB( ST value )
				{
					this->r = (float)( this->r - value );
					this->g = (float)( this->g - value );
					this->b = (float)( this->b - value );
				}						
				
				//=====================================================================================================
				/// @brief		Scalar subtraction. Subtracts this color's RGBA values by the given amount.
				/// @param		value - value to subtract by
				//=====================================================================================================
				
				template <class ST> inline void SubRGBA( ST value )
				{
					this->r = (float)( this->r - value );
					this->g = (float)( this->g - value );
					this->b = (float)( this->b - value );
					this->a = (float)( this->a - value );
				}				
				
				//=====================================================================================================
				/// @brief		Scalar multiplication. Multiplies this color's RGB values by the given amount.
				/// @param		scale - scaling amount
				//=====================================================================================================
				
				template <class ST> inline void MulRGB( ST scale )
				{
					this->r = (float)( this->r * scale );
					this->g = (float)( this->g * scale );
					this->b = (float)( this->b * scale );
				}						
				
				//=====================================================================================================
				/// @brief		Scalar multiplication. Multiplies this color's RGBA values by the given amount.
				/// @param		scale - scaling amount
				//=====================================================================================================
				
				template <class ST> inline void MulRGBA( ST scale )
				{
					this->r = (float)( this->r * scale );
					this->g = (float)( this->g * scale );
					this->b = (float)( this->b * scale );
					this->a = (float)( this->a * scale );
				}																	
				
				//=====================================================================================================
				/// @brief		Scalar division. Scales this color's RGB values by the given amount.
				/// @param		scale - scaling amount
				/// @pre		Scale must not be zero
				//=====================================================================================================
				
				template <class ST> inline void DivRGB( ST scale )
				{
					ENSURE( scale != 0 , "Scale must not be zero" );
				
					this->r = (float)( this->r / scale );
					this->g = (float)( this->g / scale );
					this->b = (float)( this->b / scale );
				}						
				
				//=====================================================================================================
				/// @brief		Scalar division. Scales this color's RGBA values by the given amount.
				/// @param		scale - scaling amount
				/// @pre		Scale must not be zero
				//=====================================================================================================
				
				template <class ST> inline void DivRGBA( ST scale )
				{
					ENSURE( scale != 0 , "Scale must not be zero" );
				
					this->r = (float)( this->r / scale );
					this->g = (float)( this->g / scale );
					this->b = (float)( this->b / scale );
					this->a = (float)( this->a / scale );
				}				
				
				//=====================================================================================================
				/// @brief		Dot product. Returns the dot of the color's RGB vector with the given 4d vector's xyz
				///				components.
				/// @param		v - 4d vector to do a dot product with
				//=====================================================================================================
				
				template <class VT> inline float DotRGB( const Vector4<VT> & v ) const
				{
					return
					(
						(float)( this->r * v.x )
						+ 
						(float)( this->g * v.y )
						+ 
						(float)( this->b * v.z )						
					);
				}					
				
				//=====================================================================================================
				/// @brief		Dot product. Returns the dot of the color's RGBA vector with the given 4d vector.
				/// @param		v - 4d vector to do a dot product with
				//=====================================================================================================
				
				template <class VT> inline float DotRGBA( const Vector4<VT> & v ) const
				{
					return
					(
						(float)( this->r * v.x )
						+ 
						(float)( this->g * v.y )
						+ 
						(float)( this->b * v.z )
						+ 
						(float)( this->a * v.w )						
					);
				}				
				
				//=====================================================================================================
				/// @brief		Dot product. Returns the dot of the color's RGB vector with the given 3d vector's
				///				xyz components.
				/// @param		v - 3d vector to do a dot product with
				//=====================================================================================================
				
				template <class VT> inline float DotRGB( const Vector3<VT> & v ) const
				{
					return
					(
						(float)( this->r * v.x )
						+ 
						(float)( this->g * v.y )
						+ 
						(float)( this->b * v.z )						
					);
				}					
				
				//=====================================================================================================
				/// @brief		Dot product. Returns the dot of the color's RGB vector with the given color's RGB
				///				vector
				/// @param		c - color to do a dot product with
				//=====================================================================================================
				
				inline float DotRGB( const Color & c ) const
				{
					return
					(
						( this->r * c.r )
						+ 
						( this->g * c.g )
						+ 
						( this->b * c.b )						
					);
				}					
				
				//=====================================================================================================
				/// @brief		Dot product. Returns the dot of the color's RGBA vector with the given color's RGBA
				///				vector
				/// @param		c - color to do a dot product with
				//=====================================================================================================
				
				inline float DotRGBA( const Color & c ) const
				{
					return
					(
						( this->r * c.r )
						+ 
						( this->g * c.g )
						+ 
						( this->b * c.b )
						+ 
						( this->a * c.a )						
					);
				}				
				
				//=====================================================================================================
				/// @brief		Gets the squared length of this color's RGBA color vector.
				/// @returns	Squared length of the color's RGBA color vector
				//=====================================================================================================
				
				inline float SqLengthRGBA() const
				{
					// Return square length:
					
					return ( this->r * this->r ) + ( this->g * this->g ) + ( this->b * this->b ) + ( this->a * this->a );
				}
				
				//=====================================================================================================
				/// @brief		Gets the squared length of this color's RGB color vector.
				/// @returns	Squared length of the color's RGB color vector
				//=====================================================================================================
				
				inline float SqLengthRGB() const
				{
					// Return square length:
					
					return ( this->r * this->r ) + ( this->g * this->g ) + ( this->b * this->b );
				}				
				
				//=====================================================================================================
				/// @brief		Gets the length of the colors RGBA color vector.
				/// @returns	Length of the color's RGBA color vector
				//=====================================================================================================
				
				inline float LengthRGBA() const
				{
					// Get square length:
					
					float sq_length = SqLengthRGBA();
					
					// Get the square root of it and return: 
					
					return sqrtf(sq_length);
				}
				
				//=====================================================================================================
				/// @brief		Gets the length of the colors RGB color vector.
				/// @returns	Length of the color's RGB color vector
				//=====================================================================================================
				
				inline float LengthRGB() const
				{
					// Get square length:
					
					float sq_length = SqLengthRGB();
					
					// Get the square root of it and return: 
					
					return sqrtf(sq_length);
				}				
				
				//=====================================================================================================
				/// @brief		Normalises the RGBA components of the color vectors color and returns the length 
				///				of the vector in the process.
				/// @returns	Length of the color vector
				/// @remarks	A check for division by zero is done
				//=====================================================================================================
				
				inline float NormalizeRGBA()
				{
					// Get length
					
					float length = LengthRGBA();
					
					// Normalize if not zero:
					
					if ( length != 0.0f )
					{
						r /= length;
						g /= length;
						b /= length;
						a /= length;
					}
					
					// Return the length of the vector
					
					return length;
				}
				
				//=====================================================================================================
				/// @brief		Normalises the RGB components of the color vectors color and returns the length 
				///				of the color's RGB vector in the process.
				/// @returns	Length of the color vector
				/// @remarks	A check for division by zero is done
				//=====================================================================================================
				
				inline float NormalizeRGB()
				{
					// Get length
					
					float length = LengthRGB();
					
					// Normalize if not zero:
					
					if ( length != 0.0f )
					{
						r /= length;
						g /= length;
						b /= length;
					}
					
					// Return the length of the vector
					
					return length;
				}				
				
				//=====================================================================================================
				/// @brief		Saturates this color, clamping all it's RGBA values to a 0-1 range.
				//=====================================================================================================
				
				inline void SaturateRGBA()
				{
					// Clamp everything:
					
					if ( r < 0 ) r = 0;
					if ( g < 0 ) g = 0;
					if ( b < 0 ) b = 0;
					if ( a < 0 ) a = 0;
					
					if ( r > 1 ) r = 1;
					if ( g > 1 ) g = 1;
					if ( b > 1 ) b = 1;
					if ( a > 1 ) a = 1;
				}
				
				//=====================================================================================================
				/// @brief		Saturates this color, clamping all it's RGB values to a 0-1 range.
				//=====================================================================================================
				
				inline void SaturateRGB()
				{
					// Clamp everything:
					
					if ( r < 0 ) r = 0;
					if ( g < 0 ) g = 0;
					if ( b < 0 ) b = 0;
					
					if ( r > 1 ) r = 1;
					if ( g > 1 ) g = 1;
					if ( b > 1 ) b = 1;
				}				
				
				//=====================================================================================================
				/// @brief		Gives a saturated version of this color with it's RGBA
				///				color components all clamped to a 0-1 range
				/// @returns	Saturated version of this color
				//=====================================================================================================
				
				inline Color SaturatedRGBA() const
				{
					// Make a copy of this color
					
					Color copy = * this;
				
					// Clamp everything:
					
					if ( copy.r < 0 ) copy.r = 0;
					if ( copy.g < 0 ) copy.g = 0;
					if ( copy.b < 0 ) copy.b = 0;
					if ( copy.a < 0 ) copy.a = 0;
					
					if ( copy.r > 1 ) copy.r = 1;
					if ( copy.g > 1 ) copy.g = 1;
					if ( copy.b > 1 ) copy.b = 1;
					if ( copy.a > 1 ) copy.a = 1;
					
					// Return the new copy
					
					return copy;
				}
				
				//=====================================================================================================
				/// @brief		Gives a saturated version of this color with it's RGB
				///				color components all clamped to a 0-1 range
				/// @returns	Saturated version of this color
				//=====================================================================================================
				
				inline Color SaturatedRGB() const
				{
					// Make a copy of this color
					
					Color copy = * this;
				
					// Clamp everything bar alpha:
					
					if ( copy.r < 0 ) copy.r = 0;
					if ( copy.g < 0 ) copy.g = 0;
					if ( copy.b < 0 ) copy.b = 0;
					
					if ( copy.r > 1 ) copy.r = 1;
					if ( copy.g > 1 ) copy.g = 1;
					if ( copy.b > 1 ) copy.b = 1;
					
					// Return the new copy
					
					return copy;
				}				
				
				//=====================================================================================================
				/// @brief		Inverts this colors RGB values. This function does not touch alpha however.
				/// @remarks	This function automatically saturates this color's RGB values.
				//=====================================================================================================				
				
				inline Color InvertRGB()
				{
					// Saturate this color:
					
					SaturateRGB();
					
					// Invert all color values
					
					r = 1.0f - r;					
					g = 1.0f - g;
					b = 1.0f - b;
				}
				
				//=====================================================================================================
				/// @brief		Inverts this colors RGBA values. This function does not touch alpha however.
				/// @remarks	This function automatically saturates this color's RGBA values.
				//=====================================================================================================				
				
				inline Color InvertRGBA()
				{
					// Saturate this color:
					
					SaturateRGBA();
					
					// Invert all color values
					
					r = 1.0f - r;					
					g = 1.0f - g;
					b = 1.0f - b;
					a = 1.0f - a;
				}											
				
				//=====================================================================================================
				/// @brief		Inverts the color. This function does not touch alpha however.
				/// @returns	A new color, which is the inverse of this color.
				/// @remarks	This function automatically saturates the new color during calculations, so data outside
				///				the 0-1 range will be lost.
				//=====================================================================================================				
				
				inline Color InverseRGB() const
				{
					// Get the saturated version of this:
					
					Color saturated = SaturatedRGB();
				
					// Return an inverted version of the color
					
					return Color
					(
						1.0f - saturated.r,
						1.0f - saturated.g,
						1.0f - saturated.b,
						saturated.a
					);
				}
				
				//=====================================================================================================
				/// @brief		Inverts the color. This function also inverts alpha.
				/// @returns	A new color, which is the inverse of this color.
				/// @remarks	This function automatically saturates the new color during calculations, so data outside
				///				the 0-1 range will be lost.
				//=====================================================================================================				
				
				inline Color InverseRGBA() const
				{
					// Get the saturated version of this:
					
					Color saturated = SaturatedRGBA();
				
					// Return an inverted version of the color
					
					return Color
					(
						1.0f - saturated.r,
						1.0f - saturated.g,
						1.0f - saturated.b,
						1.0f - saturated.a
					);
				}				
				
				//=====================================================================================================
				/// @brief		Returns the normalized version of this color, whose RGB color vector is of unit length. 
				///				If the color is black however it will stay black. 
				/// @returns	Normalized version of the vector.
				//=====================================================================================================												
				
				inline Color NormalisedRGB() const
				{
					// Make a copy of this color:
					
					Color c(*this);
					
					// Normalise:
					
					c.NormalizeRGB();
					
					// Return it:
					
					return c;
				}
				
				//=====================================================================================================
				/// @brief		Returns the normalized version of this color, whose RGBA color vector is of unit length. 
				///				If the color is black however it will stay black. 
				/// @returns	Normalized version of the vector.
				//=====================================================================================================												
				
				inline Color NormalisedRGBA() const
				{
					// Make a copy of this color:
					
					Color c(*this);
					
					// Normalise:
					
					c.NormalizeRGBA();
					
					// Return it:
					
					return c;
				}				
				
				//=====================================================================================================
				/// @brief		Tells if the two colors are exactly equal
				/// @param		c - color to compare to
				/// @returns	True if the two colors are exactly equal
				//=====================================================================================================											
				
				inline bool operator == ( const Color & c ) const
				{
					// See if equal:
					
					if 
					( 
						c.r == this->r &&
						c.g == this->g &&
						c.b == this->b &&
						c.a == this->a
					)
					{
						return true;
					}
					else
					{
						return false;
					}
				}	
				
				//=====================================================================================================
				/// @brief		Tells if the two colors are exactly not equal
				/// @param		c - color to compare to
				/// @returns	True if the two colors are exactly not equal
				//=====================================================================================================										
				
				inline bool operator != ( const Color & c ) const
				{
					// See if not equal:
					
					if 
					( 
						c.r != this->r ||
						c.g != this->g ||
						c.b != this->b ||
						c.a != this->a
					)
					{
						return true;
					}
					else
					{
						return false;
					}
				}			
				
				//=====================================================================================================
				/// @brief		Compares two colors for equality with a given margin for error. This version tests
				///				RGBA values on both vectors.
				/// @param		c - other color to compare to
				/// @param		error_margin - margin for error. this is the maximum distance in length between the two 
				///				color vectors.
				/// @remarks	This is slower than a normal compare, but can be useful for situations where a margin
				///				of error is required.
				//=====================================================================================================
				
				template<class ET> inline bool EqualsRGBA( const Color & c , ET error_margin = 0.0001f ) const
				{
					// Subtract the two vectors:
					
					Color sub =
					(
						this->r - c.r,
						this->g - c.g,
						this->b - c.b,
						this->a - c.a
					);
					
					// Get the square length of it:
					
					ET sq_length = (ET) sub.SqLengthRGBA();
					
					// See if less than or equal to error margin:
					
					if ( sq_length <= error_margin * error_margin )
					{
						// Equal within margin for error:
						
						return true;
					}
					else
					{
						// Not equal:
						
						return false;
					}
				}
				
				//=====================================================================================================
				/// @brief		Compares two colors for equality with a given margin for error. This version tests
				///				RGB values on both vectors.
				/// @param		c - other color to compare to
				/// @param		error_margin - margin for error. this is the maximum distance in length between the two 
				///				color vectors.
				/// @remarks	This is slower than a normal compare, but can be useful for situations where a margin
				///				of error is required.
				//=====================================================================================================
				
				template<class ET> inline bool EqualsRGB( const Color & c , ET error_margin = 0.0001f ) const
				{
					// Subtract the two vectors:
					
					Color sub =
					(
						this->r - c.r,
						this->g - c.g,
						this->b - c.b
					);
					
					// Get the square length of it:
					
					ET sq_length = (ET) sub.SqLengthRGB();
					
					// See if less than or equal to error margin:
					
					if ( sq_length <= error_margin * error_margin )
					{
						// Equal within margin for error:
						
						return true;
					}
					else
					{
						// Not equal:
						
						return false;
					}
				}				
				
				//=====================================================================================================
				/// @brief		Casts the color to a 4d vector
				/// @returns	A Vector4 object
				//=====================================================================================================
				
				inline Vector4<float> ToVec4() const
				{
					return Vector4<float>(r,g,b,a);
				}				
				
				//=====================================================================================================
				/// @brief		Casts the vector to a 3d vector, discarding alpha
				/// @returns	A Vector3 object
				//=====================================================================================================
				
				inline Vector3<float> ToVec3() const
				{
					return Vector3<float>(r,g,b);
				}				
				
				//=====================================================================================================
				/// @brief		Casts the color to a 2d vector, discarding blue and alpha
				/// @returns	A Vector2 object
				//=====================================================================================================
				
				inline Vector2<float> ToVec2() const
				{
					return Vector2<float>(r,g);
				}
				
				//=====================================================================================================
				/// @brief		Casts the vector to a directx vector2, discarding blue and alpha
				/// @returns	A direct x vector 2
				//=====================================================================================================				
				
				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
					inline D3DXVECTOR2 ToDxVec2() const
					{
						return D3DXVECTOR2( r , g );
					}
				
				#endif			
				
				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Casts the color to a directx vector3, discarding alpha
				/// @returns	A direct x vector 3
				//=====================================================================================================				
				
				inline D3DXVECTOR3 ToDxVec3() const
				{
					return D3DXVECTOR3( r , g , b );
				}
					
				#endif		
				
				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Casts the color to a directx vector4
				/// @returns	A direct x vector 4
				//=====================================================================================================				
				
				inline D3DXVECTOR4 ToDxVec4() const
				{
					return D3DXVECTOR4( r , g , b , a );
				}
				
				#endif
				
				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Casts the color to a directx color
				/// @returns	A direct x color
				//=====================================================================================================				
				
				inline D3DXCOLOR ToDxCol() const
				{
					return D3DXCOLOR( r , g , b , a );
				}
				
				#endif
				
				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Converts the given color into a 32bit unsigned integer, but in the directx format.			
				/// @returns	A 32-bit integer representing the color.
				/// @remarks	The hex RGBA components of the integer are laid out as follows:
				///				0xAARRGGBB. So the most significant byte is alpha and the least is blue.
				///				The color is automatically saturated by this function to clamp to a 0-1 range.
				//=====================================================================================================
				
				inline unsigned ToDxUnsigned() const
				{								
					// Get a saturated version
					
					Color s = SaturatedRGBA();
					
					// Inflate RGB components
					
					float rf = float(s.r) * float(255.0f);
					float gf = float(s.g) * float(255.0f);
					float bf = float(s.b) * float(255.0f);
					float af = float(s.a) * float(255.0f);
					
					// Convert to integer and return:
					
					return 
					
						(((unsigned)(af)) << 24 ) |
						(((unsigned)(rf)) << 16 ) |
						(((unsigned)(gf)) << 8  ) |
						(((unsigned)(bf))		) ;
				}						
					
				#endif
				
				//=====================================================================================================
				/// @brief		Converts the color into array format, saving it to the given array
				/// @param		a - array to save to
				/// @pre		Given array cannot be null
				//=====================================================================================================
				
				inline void ToArray( float * a ) const
				{
					ENSURE(a,"Given array cannot be null");
					
					a[0] = this->r;
					a[1] = this->g;
					a[2] = this->b;
					a[3] = this->a;
				}				
				
				//=====================================================================================================
				/// @brief		Converts the given color into a 32bit unsigned integer.				
				/// @returns	A 32-bit integer representing the color.
				/// @remarks	The hex RGBA components of the integer are laid out as follows:
				///				0xRRGGBBAA. So the most significant byte is red and the least is alpha.
				///				The color is automatically saturated by this function to clamp to a 0-1 range.
				//=====================================================================================================
				
				inline unsigned ToUnsigned() const
				{								
					// Get a saturated version
					
					Color s = SaturatedRGBA();
					
					// Inflate RGB components
					
					float rf = float(s.r) * float(255.0f);
					float gf = float(s.g) * float(255.0f);
					float bf = float(s.b) * float(255.0f);
					float af = float(s.a) * float(255.0f);
					
					// Convert to integer and return:
					
					return 
					
						(((unsigned)(rf)) << 24 ) |
						(((unsigned)(gf)) << 16 ) |
						(((unsigned)(bf)) << 8  ) |
						(((unsigned)(af))		) ;
				}				
				
				//=====================================================================================================
				/// @brief		Returns opaque red
				/// @returns	Opaque red
				//=====================================================================================================
				
				static inline Color Red(){ return Color(1.0f,0.0f,0.0f,1.0f); }
				
				//=====================================================================================================
				/// @brief		Returns opaque green
				/// @returns	Opaque red
				//=====================================================================================================
				
				static inline Color Green(){ return Color(0.0f,1.0f,0.0f,1.0f); }
				
				//=====================================================================================================
				/// @brief		Returns opaque blue
				/// @returns	Opaque blue
				//=====================================================================================================
				
				static inline Color Blue(){ return Color(0.0f,0.0f,1.0f,1.0f); }
				
				//=====================================================================================================
				/// @brief		Returns opaque white
				/// @returns	Opaque white
				//=====================================================================================================
				
				static inline Color White(){ return Color(1.0f,1.0f,1.0f,1.0f); }
				
				//=====================================================================================================
				/// @brief		Returns opaque black
				/// @returns	Opaque white
				//=====================================================================================================
				
				static inline Color Black(){ return Color(0.0f,0.0f,0.0f,1.0f); }
				
				//=====================================================================================================
				/// @brief		Returns opaque cyan
				/// @returns	Opaque cyan
				//=====================================================================================================
				
				static inline Color Cyan(){ return Color(0.0f,1.0f,1.0f,1.0f); }
				
				//=====================================================================================================
				/// @brief		Returns opaque magenta
				/// @returns	Opaque magenta
				//=====================================================================================================
				
				static inline Color Magenta(){ return Color(1.0f,0.0f,1.0f,1.0f); }
				
				//=====================================================================================================
				/// @brief		Returns opaque yellow
				/// @returns	Opaque yellow
				//=====================================================================================================
				
				static inline Color Yellow(){ return Color(1.0f,1.0f,0.0f,1.0f); }				
				
				//=====================================================================================================
				/// @brief		Returns a transparent black color
				/// @returns	Transparent black
				//=====================================================================================================
				
				static inline Color TransparentBlack(){ return Color(0.0f,0.0f,0.0f,1.0f); }									
		
			//=========================================================================================================
			// Variables
			//=========================================================================================================
			
				/// Red component for the vector
				
				float r;
				
				/// Green component for the vector
				
				float g;

				/// Blue component for the vector
				
				float b;
				
				/// Alpha component for the vector
				
				float a;					
	};
};

#endif 
