/*
	Trig Wars.
	Copyright (c) 2008, Darragh Coy.
	All rights reserved.
	Original work(s), source code & author(s) details available at: http://www.darraghcoy.com

	Redistribution and use in source and binary forms, with or without modification, are permitted 
	provided that the following conditions are met:

		* 	Redistributions of source code must retain the above copyright notice, this list of 
			conditions and the following disclaimer.
		* 	Redistributions in binary form must reproduce the above copyright notice, this list of 
			conditions and the following disclaimer in the documentation.
		* 	Neither the name of the author nor the names of contributors may be used to endorse or 
			promote products derived from this software without specific prior written permission.
		* 	The use of any third party libraries, modules or data files provided with the source 
			form of the program must follow and respect the individual licensing terms of these 
			components. You do not have permission to modify or circumvent the licensing terms for 
			provided third party components unless the third party licenses specifically allow this. 
			If you do not agree to the terms of provided third party components then you do not have 
			permission to use them in your derived work(s).
		* 	The terms of this license do not apply to separate (but related) distributions if supplied 
			by the author(s), only to the contents of this source distribution, derived works, and the 
			binaries produced by either modified or unmodified versions of the source code. The terms 
			of this license do not also apply to any third party libraries or components provided with 
			the source distribution. Such third party elements are licensed under their own separate 
			agreements.

	THIS SOFTWARE IS PROVIDED BY DARRAGH COY ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
	BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL DARRAGH COY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
	IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef GAME_MATRIX2_H
#define GAME_MATRIX2_H

//=====================================================================================================================
//=====================================================================================================================

#include "Base\Assertion.h"
#include "Base\Math\Float.h"
#include "Base\Math\Vector2.h"

// Specific:

#ifdef GAME_SUPPORT_DIRECT_X_TYPES

	#include <D3DX10Math.h>

#endif

//=====================================================================================================================
//=====================================================================================================================

namespace Game
{
	//=================================================================================================================
	/// @brief		Represents a 2x2 matrix
	//=================================================================================================================

	template <class T> struct Matrix2
	{
		public:

			//=========================================================================================================
			// Functions
			//=========================================================================================================

				//=====================================================================================================
				/// @brief		Constructor. Creates the identity matrix.
				//=====================================================================================================
				
				inline Matrix2()
				{	
					m_components[0][0]=(T)1; m_components[0][1]=(T)0;
					m_components[1][0]=(T)0; m_components[1][1]=(T)1;
				}

				//=====================================================================================================
				/// @brief		Constructor. Creates the matrix from the given elements.
				///	@param		r1c1 - 1st row 1st column
				///	@param		r1c2 - 1st row 2nd column
				///	@param		r2c1 - 2nd row 1st column
				///	@param		r2c2 - 2nd row 2nd column
				//=====================================================================================================
				
				template <class ET> inline Matrix2
				(
					ET r1c1,
					ET r1c2,
					ET r2c1,
					ET r2c2
				)
				{
					this->m_components[0][0] = (T) r1c1;
					this->m_components[0][1] = (T) r1c2;

					this->m_components[1][0] = (T) r2c1;
					this->m_components[1][1] = (T) r2c2;
				}

				//=====================================================================================================
				/// @brief		Constructor. Creates the matrix from a given array.
				///				Array must be row major. I.E the 1st row is stored 
				///				first in the array, second row next and so forth.
				///	@param		a - array to create from
				/// @pre		Given array cannot be null
				//=====================================================================================================
				
				template <class VT> inline Matrix2( VT * a )
				{
					// Make sure ok:

					ENSURE(a,"Given array cannot be null");

					// Do the copy:

					for ( int i = 0 ; i < 4 ; ++i )
					{
						*((T*)(m_components) + i ) = (T)(*( a + i ));
					}					
				}

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Constructor. Creates the matrix from a given DirectX 4x4 matrix, discarding anything
				///				that is not part of a the 2x2 section of it
				///	@param		m - matrix to create from
				//=====================================================================================================				

				inline Matrix2( const D3DXMATRIX & m )
				{
					// Copy it:

					for ( int r = 0; r < 2 ; r++ )
					for ( int c = 0; c < 2 ; c++ )
					{
						m_components[r][c] = m(r,c);
					}
				}

				#endif

				//=====================================================================================================
				/// @brief		Constructor. Creates the matrix from 2 column vectors.
				///	@param		a - array of column vectors
				/// @pre		Given array cannot be null
				//=====================================================================================================
				
				template <class VT> inline Matrix2( Vector2<VT> * a )
				{
					// Make sure ok:

					ENSURE(a,"Given array cannot be null");

					// Assign:

					for ( int i = 0 ; i < 2 ; ++i )
					{
						m_components[0][i] = (T) a[i].x;
						m_components[1][i] = (T) a[i].y;
					}
				}

				//=====================================================================================================
				/// @brief		Constructor. Creates the matrix from 2 column vectors.
				///	@param		c1 - column vector 1
				///	@param		c2 - column vector 2
				//=====================================================================================================
				
				template <class VT> inline Matrix2
				( 
					const Vector2<VT> & c1,
					const Vector2<VT> & c2
				)
				{
					m_components[0][0] = (T) c1.x;
					m_components[0][1] = (T) c2.x;

					m_components[1][0] = (T) c1.y;
					m_components[1][1] = (T) c2.y;
				}

				//=====================================================================================================
				/// @brief		Copy constructor. Makes this matrix from another
				/// @param		other - other matrix to assign from
				//=====================================================================================================

				template <class MT> inline Matrix2( const Matrix2<MT> & other )
				{
					// Do the copy:

					for ( int i = 0 ; i < 4 ; ++i )
					{
						*((T*)(m_components) + i ) = (T)(*((MT*)(other.m_components) + i));
					}
				}

				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the other matrix to this matrix.
				/// @param		other - other matrix to assign from
				/// @returns	This matrix
				//=====================================================================================================

				template <class MT> inline Matrix2<T> & operator = ( const Matrix2<MT> & other )
				{
					// Do the copy:

					for ( int i = 0 ; i < 4 ; ++i )
					{
						*((T*)(m_components) + i ) = (T)(*((MT*)(other.m_components) + i));
					}

					return * this;
				}

				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the array of column vectors to the matrix
				///	@param		a - array representing a matrix through column vectors
				/// @returns	This matrix
				//=====================================================================================================
				
				template <class VT> inline Matrix2<T> & operator = ( const Vector2<VT> * a )
				{
					// Make sure ok:

					ENSURE(a,"Given array cannot be null");

					// Assign:

					for ( int i = 0 ; i < 2 ; ++i )
					{
						m_components[0][i] = (T) a[i].x;
						m_components[1][i] = (T) a[i].y;
					}

					return * this;
				}

				//=====================================================================================================
				/// @brief		Assignment operator. Creates the matrix from a given array.
				///				Array must be row major. I.E the 1st row is stored 
				///				first in the array, second row next and so forth.
				///	@param		a - array to create from
				/// @pre		Given array cannot be null
				//=====================================================================================================
				
				template <class VT> inline Matrix2<T> & operator = ( VT * a )
				{
					// Make sure ok:

					ENSURE(a,"Given array cannot be null");

					// Do the copy:

					for ( int i = 0 ; i < 4 ; ++i )
					{
						*((T*)(m_components) + i ) = (T)(*( a + i ));
					}		

					return * this;
				}

				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the given 4x4 directx matrix to this matrix.
				///				This discards anything that is not part of the 2x2 part of the direct x matrix.
				///	@param		m - matrix to assign to this matrix
				/// @returns	this
				//=====================================================================================================

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES

				inline Matrix2<T> & operator = ( const D3DXMATRIX & m )
				{
					// Copy it:

					for ( int r = 0; r < 2 ; r++ )
					for ( int c = 0; c < 2 ; c++ )
					{
						m_components[r][c] = m(r,c);
					}

					return * this;
				}

				#endif

				//=====================================================================================================
				/// @brief		Matrix multiplication. Multiplies this matrix with another matrix.
				/// @param		m - matrix to multiply with
				/// @returns	a new matrix
				//=====================================================================================================

				template <class MT> inline Matrix2<T> & operator * ( const Matrix2<MT> & m ) const
				{
					// Need to make a temporary copy of this matrix:

					Matrix new_matrix;

					// Do this to shorten things:

					T  * a = (T* ) this->m_components;					
					MT * b = (MT*) m.m_components
					T  * c = (T* ) new_matrix.m_components;
					
					// Right: on to the multiply

					c[0][0] = (T)( a[0] * b[0] + a[1] * b[2] ); 
					c[0][1] = (T)( a[0] * b[1] + a[1] * b[3] );

					c[1][0] = (T)( a[2] * b[0] + a[3] * b[2] ); 
					c[1][1] = (T)( a[2] * b[1] + a[3] * b[3] );

					// Return the new matrix

					return new_matrix;
				}

				//=====================================================================================================
				/// @brief		Matrix multiplication. Multiplies this matrix with another matrix.
				/// @param		m - matrix to multiply with
				/// @returns	this matrix
				//=====================================================================================================

				template <class MT> inline void operator *= ( const Matrix2<MT> & m )
				{
					// Need to make a temporary copy of this matrix:

					Matrix temp(*this);

					// Do this to shorten things:

					T  * a = (T* ) temp.m_components;					
					MT * b = (MT*) m.m_components
					T  * c = (T* ) m_components;
					
					// Right: on to the multiply

					c[0][0] = (T)( a[0] * b[0] + a[1] * b[2] ); 
					c[0][1] = (T)( a[0] * b[1] + a[1] * b[3] );

					c[1][0] = (T)( a[2] * b[0] + a[3] * b[2] ); 
					c[1][1] = (T)( a[2] * b[1] + a[3] * b[3] );
				}

				//=====================================================================================================
				/// @brief		Retrieves the given row vector.
				/// @param		r - row to retrieve.
				/// @pre		Row index must be in range
				//======================================================================================================

				inline Vector2<T> GetRow( unsigned r ) const
				{
					ENSURE( r < 2 , "Row index must be in range" );

					return Vector2<T>
					(
						m_components[r][0],
						m_components[r][1]
					);
				}

				//=====================================================================================================
				/// @brief		Sets the given row to the given value
				/// @param		r - row to set
				/// @param		v - vector representing the row
				/// @pre		Row index must be in range
				//======================================================================================================

				template<class VT> inline void SetRow( unsigned r , const Vector2<VT> & v )
				{
					ENSURE( r < 2 , "Row index must be in range" );

					m_components[r][0] = (T) v.x;
					m_components[r][1] = (T) v.y;
				}

				//=====================================================================================================
				/// @brief		Retrieves the given column vector.
				/// @param		c - column to retrieve.
				/// @pre		Column index must be in range
				//======================================================================================================

				inline Vector2<T> GetCol( unsigned c ) const
				{
					ENSURE( c < 2 , "Column index must be in range" );

					return Vector2<T>
					(
						m_components[0][c],
						m_components[1][c]
					);
				}

				//=====================================================================================================
				/// @brief		Sets the given column to the given value
				/// @param		c - column to set
				/// @param		v - vector representing the row
				/// @pre		Column index must be in range
				//======================================================================================================

				template<class VT> inline void SetCol( unsigned c , const Vector3<VT> & v )
				{
					ENSURE( c < 2 , "Column index must be in range" );

					m_components[0][c] = (T) v.x;
					m_components[1][c] = (T) v.y;
				}

				//=====================================================================================================
				/// @brief		Saves the matrix to a given array. Matrix is saved in row major format.
				/// @param		a - array to save to.
				/// @pre		Given array cannot be null
				//=====================================================================================================

				template <class AT> inline void ToArray( AT * a ) const
				{
					// Save:

					for ( int i = 0 ; i < 4 ; ++i )
					{
						* (a + i) = (AT) * (((T*)(m_components)) + i );
					}
				}

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Saves the matrix to a given directx matrix. The missing components are set to the 
				///				identity since this is a 2x2 matrix.
				/// @param		m - matrix to save to
				//=====================================================================================================				

				inline void ToDxMatrix( D3DXMATRIX & m ) const
				{
					// Copy it:

					for ( int r = 0; r < 2 ; r++ )
					for ( int c = 0; c < 2 ; c++ )
					{
						m(r,c) = m_components[r][c];
					}	

					// Set the components we are not using:

					m(0,2) = 0;
					m(0,3) = 0;

					m(2,0) = 0;
					m(2,1) = 0;
					m(2,2) = 1;
					m(2,3) = 0;

					m(3,0) = 0;
					m(3,1) = 0;
					m(3,2) = 0;
					m(3,3) = 1;
				}

				#endif

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Saves the matrix to a given directx matrix. The missing components are set to the 
				///				identity since this is a 2x2 matrix.
				/// @returns	DirectX matrix
				//=====================================================================================================				

				inline D3DXMATRIX ToDxMatrix() const
				{
					// Make it here:

					D3DXMATRIX m;

					// Copy it:

					for ( int r = 0; r < 2 ; r++ )
					for ( int c = 0; c < 2 ; c++ )
					{
						m(r,c) = m_components[r][c];
					}	

					// Set the components we are not using:

					m(0,2) = 0;
					m(0,3) = 0;

					m(2,0) = 0;
					m(2,1) = 0;
					m(2,2) = 1;
					m(2,3) = 0;

					m(3,0) = 0;
					m(3,1) = 0;
					m(3,2) = 0;
					m(3,3) = 1;

					// Return it:

					return m;
				}

				#endif

				//=====================================================================================================
				/// @brief		Gives the identity matrix
				/// @returns	The identity matrix
				//=====================================================================================================

				static inline Matrix2<T> Identity(){ return Matrix2<T>(); }

				//=====================================================================================================
				/// @brief		Transposes the matrix and returns the answer
				/// @returns	A new matrix which is the transpose of this matrix
				//=====================================================================================================

				static inline Matrix2<T> Transpose()
				{
					// Return the new matrix:

					return Matrix2<T>
					(
						m_components[0][0],m_components[1][0],
						m_components[0][1],m_components[1][1]
					);
				}

				//=====================================================================================================
				/// @brief		Creates a rotation matrix to rotate in 2d
				/// @param		r - amount of rotation
				/// @returns	The given transform
				//=====================================================================================================

				template <class VT> static inline Matrix2<T> Rotation( VT r )
				{
					// Get cos and sin of the angle:

					float c = Float::Cos(r);
					float s = Float::Sin(r);

					// Return the matrix:

					return Matrix2<T>
					(
						(T) c	, (T) s	 ,
						(T) -s	, (T) c
					);
				}

				//=====================================================================================================
				/// @brief		Creates a scaling transform, in either 1d, or 2d
				/// @param		x - amount to scale on x axis
				/// @param		y - amount to scale on y axis
				/// @returns	The given scaling transform
				//=====================================================================================================

				template <class ST> static inline Matrix2<T> Scale( ST x , ST y = 1 )
				{ 
					return Matrix2<T>
					(
						(T) x , (T) 0 ,
						(T) 0 , (T) y
					);
				}

				//=====================================================================================================
				/// @brief		Transforms the given vector and returns the answer.
				/// @param		v - vector to transform
				/// @returns	The vector transformed by the matrix
				/// @remarks	The z and w components remain unmodified. This function treats the given vector as 
				///				if it were 2D only.
				//=====================================================================================================

				template <class VT> inline Vector4<VT> Transform( const Vector4<VT> & v ) const
				{ 
					// Save the matrix here:

					return Vector4<VT>
					(
						(VT)( v.x * m_components[0][0] + v.y * m_components[1][0] ),
						(VT)( v.x * m_components[0][1] + v.y * m_components[1][1] ),
						v.z,
						v.w
					);
				}

				//=====================================================================================================
				/// @brief		Transforms the given vector and returns the answer.
				/// @param		v - vector to transform
				/// @returns	The vector transformed by the matrix
				/// @remarks	The z component remains unmodified. This function treats the given vector as if it were
				///				2D only.
				//=====================================================================================================

				template <class VT> inline Vector3<VT> Transform( const Vector3<VT> & v ) const
				{ 
					// Save the matrix here:

					return Vector3<VT>
					(
						(VT)( v.x * m_components[0][0] + v.y * m_components[1][0] ),
						(VT)( v.x * m_components[0][1] + v.y * m_components[1][1] ),
						v.z
					);
				}

				//=====================================================================================================
				/// @brief		Transforms the given vector and returns the answer.
				/// @param		v - vector to transform
				/// @returns	The vector transformed by the matrix
				//=====================================================================================================

				template <class VT> inline Vector2<VT> Transform( const Vector2<VT> & v ) const
				{ 
					// Save the matrix here:

					return Vector2<VT>
					(
						(VT)( v.x * m_components[0][0] + v.y * m_components[1][0] ) ,
						(VT)( v.x * m_components[0][1] + v.y * m_components[1][1] )
					);
				}
		
			//=========================================================================================================
			// Variables
			//=========================================================================================================
			
				/// The components of the matrix

				T m_components[2][2];
	};
};

#endif 
