/*
	Trig Wars.
	Copyright (c) 2008, Darragh Coy.
	All rights reserved.
	Original work(s), source code & author(s) details available at: http://www.darraghcoy.com

	Redistribution and use in source and binary forms, with or without modification, are permitted 
	provided that the following conditions are met:

		* 	Redistributions of source code must retain the above copyright notice, this list of 
			conditions and the following disclaimer.
		* 	Redistributions in binary form must reproduce the above copyright notice, this list of 
			conditions and the following disclaimer in the documentation.
		* 	Neither the name of the author nor the names of contributors may be used to endorse or 
			promote products derived from this software without specific prior written permission.
		* 	The use of any third party libraries, modules or data files provided with the source 
			form of the program must follow and respect the individual licensing terms of these 
			components. You do not have permission to modify or circumvent the licensing terms for 
			provided third party components unless the third party licenses specifically allow this. 
			If you do not agree to the terms of provided third party components then you do not have 
			permission to use them in your derived work(s).
		* 	The terms of this license do not apply to separate (but related) distributions if supplied 
			by the author(s), only to the contents of this source distribution, derived works, and the 
			binaries produced by either modified or unmodified versions of the source code. The terms 
			of this license do not also apply to any third party libraries or components provided with 
			the source distribution. Such third party elements are licensed under their own separate 
			agreements.

	THIS SOFTWARE IS PROVIDED BY DARRAGH COY ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
	BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL DARRAGH COY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
	IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef GAME_MATRIX3_H
#define GAME_MATRIX3_H

//=====================================================================================================================
//=====================================================================================================================

#include "Base\Assertion.h"
#include "Base\Math\Float.h"
#include "Base\Math\Matrix2.h"
#include "Base\Math\Vector3.h"

// Specific:

#ifdef GAME_SUPPORT_DIRECT_X_TYPES

	#include <D3DX10Math.h>

#endif

//=====================================================================================================================
//=====================================================================================================================

namespace Game
{
	//=================================================================================================================
	/// @brief		Represents a 3x3 matrix
	//=================================================================================================================

	template <class T> struct Matrix3
	{
		public:

			//=========================================================================================================
			// Functions
			//=========================================================================================================

				//=====================================================================================================
				/// @brief		Constructor. Creates the identity matrix.
				//=====================================================================================================
				
				inline Matrix3()
				{
					m_components[0][0]=(T)1; m_components[0][1]=(T)0; m_components[0][2]=(T)0;
					m_components[1][0]=(T)0; m_components[1][1]=(T)1; m_components[1][2]=(T)0;
					m_components[2][0]=(T)0; m_components[2][1]=(T)0; m_components[2][2]=(T)1;
				}

				//=====================================================================================================
				/// @brief		Constructor. Creates from a 2x2 matrix and assigns missing components to the identity
				/// @param		m - matrix to construct from
				//=====================================================================================================
				
				template<class MT> inline Matrix3( Matrix2<MT> & m )
				{
					m_components[0][0]=(T) m.m_components[0][0];
					m_components[0][1]=(T) m.m_components[0][1];
					m_components[0][2]=(T) 0;

					m_components[1][0]=(T) m.m_components[1][0];
					m_components[1][1]=(T) m.m_components[1][1];
					m_components[1][2]=(T) 0;

					m_components[2][0]=(T) 0;
					m_components[2][1]=(T) 0;
					m_components[2][2]=(T) 1;
				}

				//=====================================================================================================
				/// @brief		Constructor. Creates the matrix from the given elements.
				///	@param		r1c1 - 1st row 1st column
				///	@param		r1c2 - 1st row 2nd column
				///	@param		r1c3 - 1st row 3rd column
				///	@param		r2c1 - 2nd row 1st column
				///	@param		r2c2 - 2nd row 2nd column
				///	@param		r2c3 - 2nd row 3rd column
				///	@param		r3c1 - 3rd row 1st column
				///	@param		r3c2 - 3rd row 3rd column
				///	@param		r3c3 - 3rd row 3rd column
				//=====================================================================================================
				
				template <class ET> inline Matrix3
				(
					ET r1c1,
					ET r1c2,
					ET r1c3,
					ET r2c1,
					ET r2c2,
					ET r2c3,
					ET r3c1,
					ET r3c2,
					ET r3c3
				)
				{
					this->m_components[0][0] = (T) r1c1;
					this->m_components[0][1] = (T) r1c2;
					this->m_components[0][2] = (T) r1c3;

					this->m_components[1][0] = (T) r2c1;
					this->m_components[1][1] = (T) r2c2;
					this->m_components[1][2] = (T) r2c3;

					this->m_components[2][0] = (T) r3c1;
					this->m_components[2][1] = (T) r3c2;
					this->m_components[2][2] = (T) r3c3;
				}

				//=====================================================================================================
				/// @brief		Constructor. Creates the matrix from a given array.
				///				Array must be row major. I.E the 1st row is stored 
				///				first in the array, second row next and so forth.
				///	@param		a - array to create from
				/// @pre		Given array cannot be null
				//=====================================================================================================
				
				template <class VT> inline Matrix3( VT * a )
				{
					// Make sure ok:

					ENSURE(a,"Given array cannot be null");

					// Do the copy:

					for ( int i = 0 ; i < 9 ; ++i )
					{
						*((T*)(m_components) + i ) = (T)(*( a + i ));
					}					
				}
				
				#ifdef GAME_SUPPORT_DIRECT_X_TYPES

				//=====================================================================================================
				/// @brief		Constructor. Creates the matrix from a given DirectX 4x4 matrix, discarding the 
				///				translational part of the matrix.
				///	@param		m - matrix to create from
				//=====================================================================================================

				inline Matrix3( const D3DXMATRIX & m )
				{
					// Copy it:

					for ( int r = 0; r < 3 ; r++ )
					for ( int c = 0; c < 3 ; c++ )
					{
						m_components[r][c] = m(r,c);
					}
				}

				#endif

				//=====================================================================================================
				/// @brief		Constructor. Creates the matrix from 3 column vectors.
				///	@param		a - array of column vectors
				/// @pre		Given array cannot be null
				//=====================================================================================================
				
				template <class VT> inline Matrix3( Vector3<VT> * a )
				{
					// Make sure ok:

					ENSURE(a,"Given array cannot be null");

					// Assign:

					for ( int i = 0 ; i < 3 ; ++i )
					{
						m_components[0][i] = (T) a[i].x;
						m_components[1][i] = (T) a[i].y;
						m_components[2][i] = (T) a[i].z;
					}
				}

				//=====================================================================================================
				/// @brief		Constructor. Creates the matrix from 3 column vectors.
				///	@param		c1 - column vector 1
				///	@param		c2 - column vector 2
				///	@param		c3 - column vector 3
				//=====================================================================================================
				
				template <class VT> inline Matrix3
				( 
					const Vector3<VT> & c1,
					const Vector3<VT> & c2,
					const Vector3<VT> & c3
				)
				{
					m_components[0][0] = (T) c1.x;
					m_components[0][1] = (T) c2.x;
					m_components[0][2] = (T) c3.x;

					m_components[1][0] = (T) c1.y;
					m_components[1][1] = (T) c2.y;
					m_components[1][2] = (T) c3.y;

					m_components[2][0] = (T) c1.z;
					m_components[2][1] = (T) c2.z;
					m_components[2][2] = (T) c3.z;
				}

				//=====================================================================================================
				/// @brief		Copy constructor. Makes this matrix from another
				/// @param		other - other matrix to assign from
				//=====================================================================================================

				template <class MT> inline Matrix3( const Matrix3<MT> & other )
				{
					// Do the copy:

					for ( int i = 0 ; i < 9 ; ++i )
					{
						*((T*)(m_components) + i ) = (T)(*((MT*)(other.m_components) + i));
					}
				}

				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the other matrix to this matrix. Makes up for the other 
				///				missing components by assigning them the identity.
				/// @param		other - other matrix to assign from
				/// @returns	This matrix
				//=====================================================================================================
				
				template <class MT> inline Matrix3<T> & operator = ( const Matrix2<MT> & other )
				{
					m_components[0][0]=(T) other.m_components[0][0];
					m_components[0][1]=(T) other.m_components[0][1];
					m_components[0][2]=(T) 0;

					m_components[1][0]=(T) other.m_components[1][0];
					m_components[1][1]=(T) other.m_components[1][1];
					m_components[1][2]=(T) 0;

					m_components[2][0]=(T) 0;
					m_components[2][1]=(T) 0;
					m_components[2][2]=(T) 1;

					return * this;
				}

				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the other matrix to this matrix.
				/// @param		other - other matrix to assign from
				/// @returns	This matrix
				//=====================================================================================================

				template <class MT> inline Matrix3<T> & operator = ( const Matrix3<MT> & other )
				{
					// Do the copy:

					for ( int i = 0 ; i < 9 ; ++i )
					{
						*((T*)(m_components) + i ) = (T)(*((MT*)(other.m_components) + i));
					}

					return * this;
				}

				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the given array of column vectors to the matrix
				///	@param		a - array representing a matrix
				/// @returns	This matrix
				/// @pre		Given array cannot be null	
				//=====================================================================================================
				
				template <class VT> inline Matrix3<T> & operator = ( const Vector3<VT> * a )
				{
					// Make sure ok:

					ENSURE(a,"Given array cannot be null");

					// Assign:

					for ( int i = 0 ; i < 3 ; ++i )
					{
						m_components[0][i] = (T) a[i].x;
						m_components[1][i] = (T) a[i].y;
						m_components[2][i] = (T) a[i].z;
					}

					return * this;
				}

				//=====================================================================================================
				/// @brief		Assignment operator. Creates the matrix from a given array.
				///				Array must be row major. I.E the 1st row is stored 
				///				first in the array, second row next and so forth.
				///	@param		a - array to create from
				/// @returns	This matrix
				/// @pre		Given array cannot be null				
				//=====================================================================================================
				
				template <class VT> inline Matrix3<T> & operator = ( VT * a )
				{
					// Make sure ok:

					ENSURE(a,"Given array cannot be null");

					// Do the copy:

					for ( int i = 0 ; i < 9 ; ++i )
					{
						*((T*)(m_components) + i ) = (T)(*( a + i ));
					}		

					return * this;
				}

				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the given 4x4 directx matrix to this matrix.
				///	@param		m - matrix to assign to this matrix
				/// @returns	this
				//=====================================================================================================

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES

					inline Matrix3<T> & operator = ( const D3DXMATRIX & m )
					{
						// Copy it:

						for ( int r = 0; r < 3 ; r++ )
						for ( int c = 0; c < 3 ; c++ )
						{
							m_components[r][c] = m(r,c);
						}

						return * this;
					}

				#endif

				//=====================================================================================================
				/// @brief		Matrix multiplication. Multiplies this matrix with another matrix.
				/// @param		m - matrix to multiply with
				/// @returns	a new matrix
				//=====================================================================================================

				template <class MT> inline Matrix3<T> operator * ( const Matrix3<MT> & m ) const
				{
					// Need to make a temporary copy of this matrix:

					Matrix new_matrix;

					// Do this to shorten things:

					T  * a = (T* ) this->m_components;					
					MT * b = (MT*) m.m_components
					T  * c = (T* ) new_matrix.m_components;
					
					// Right: on to the multiply

					c[0][0] = (T)( a[0] * b[0] + a[1] * b[3] + a[2] * b[6] ); 
					c[0][1] = (T)( a[0] * b[1] + a[1] * b[4] + a[2] * b[7] );
					c[0][2] = (T)( a[0] * b[2] + a[1] * b[5] + a[2] * b[8] );

					c[1][0] = (T)( a[3] * b[0] + a[4] * b[3] + a[5] * b[6] ); 
					c[1][1] = (T)( a[3] * b[1] + a[4] * b[4] + a[5] * b[7] );
					c[1][2] = (T)( a[3] * b[2] + a[4] * b[5] + a[5] * b[8] );

					c[2][0] = (T)( a[6] * b[0] + a[7] * b[3] + a[8] * b[6] ); 
					c[2][1] = (T)( a[6] * b[1] + a[7] * b[4] + a[8] * b[7] );
					c[2][2] = (T)( a[6] * b[2] + a[7] * b[5] + a[8] * b[8] );

					// Return the new matrix

					return new_matrix;
				}

				//=====================================================================================================
				/// @brief		Matrix multiplication. Multiplies this matrix with another matrix.
				/// @param		m - matrix to multiply with
				/// @returns	this matrix
				//=====================================================================================================

				template <class MT> inline void operator *= ( const Matrix3<MT> & m )
				{
					// Need to make a temporary copy of this matrix:

					Matrix temp(*this);

					// Do this to shorten things:

					T  * a = (T* ) temp.m_components;					
					MT * b = (MT*) m.m_components
					T  * c = (T* ) m_components;
					
					// Right: on to the multiply

					c[0][0] = (T)( a[0] * b[0] + a[1] * b[3] + a[2] * b[6] ); 
					c[0][1] = (T)( a[0] * b[1] + a[1] * b[4] + a[2] * b[7] );
					c[0][2] = (T)( a[0] * b[2] + a[1] * b[5] + a[2] * b[8] );

					c[1][0] = (T)( a[3] * b[0] + a[4] * b[3] + a[5] * b[6] ); 
					c[1][1] = (T)( a[3] * b[1] + a[4] * b[4] + a[5] * b[7] );
					c[1][2] = (T)( a[3] * b[2] + a[4] * b[5] + a[5] * b[8] );

					c[2][0] = (T)( a[6] * b[0] + a[7] * b[3] + a[8] * b[6] ); 
					c[2][1] = (T)( a[6] * b[1] + a[7] * b[4] + a[8] * b[7] );
					c[2][2] = (T)( a[6] * b[2] + a[7] * b[5] + a[8] * b[8] );
				}

				//=====================================================================================================
				/// @brief		Retrieves the given row vector.
				/// @param		r - row to retrieve.
				/// @pre		Row index must be in range
				//======================================================================================================

				inline Vector3<T> GetRow( unsigned r ) const
				{
					ENSURE( r < 3 , "Row index must be in range" );

					return Vector3<T>
					(
						m_components[r][0],
						m_components[r][1],
						m_components[r][2],
					);
				}

				//=====================================================================================================
				/// @brief		Sets the given row to the given value
				/// @param		r - row to set
				/// @param		v - vector representing the row
				/// @pre		Row index must be in range
				//======================================================================================================

				template<class VT> inline void SetRow( unsigned r , const Vector3<VT> & v )
				{
					ENSURE( r < 3 , "Row index must be in range" );

					m_components[r][0] = (T) v.x;
					m_components[r][1] = (T) v.y;
					m_components[r][2] = (T) v.z;
				}

				//=====================================================================================================
				/// @brief		Retrieves the given column vector.
				/// @param		c - column to retrieve.
				/// @pre		Column index must be in range
				//======================================================================================================

				inline Vector3<T> GetCol( unsigned c ) const
				{
					ENSURE( c < 3 , "Column index must be in range" );

					return Vector3<T>
					(
						m_components[0][c],
						m_components[1][c],
						m_components[2][c],
					);
				}

				//=====================================================================================================
				/// @brief		Sets the given column to the given value
				/// @param		c - column to set
				/// @param		v - vector representing the row
				/// @pre		Column index must be in range
				//======================================================================================================

				template<class VT> inline void SetCol( unsigned c , const Vector3<VT> & v )
				{
					ENSURE( c < 3 , "Column index must be in range" );

					m_components[0][c] = (T) v.x;
					m_components[1][c] = (T) v.y;
					m_components[2][c] = (T) v.z;
				}

				//=====================================================================================================
				/// @brief		Saves the matrix to a given array. Matrix is saved in row major format.
				/// @param		a - array to save to.
				/// @pre		Given array cannot be null
				//=====================================================================================================

				template <class AT> inline void ToArray( AT * a ) const
				{
					// Save:

					for ( int i = 0 ; i < 9 ; ++i )
					{
						* (a + i) = (AT) * (((T*)(m_components)) + i );
					}
				}

				//=====================================================================================================
				/// @brief		Converts to a 2x2 matrix by discarding the extra 3x3 bits
				/// @returns	This matrix in it's 2x2 form
				//=====================================================================================================

				inline void ToMatrix2() const
				{
					return Matrix2<T>
					(
						m_components[0][0],
						m_components[0][1],
						m_components[1][0],
						m_components[1][1]
					);
				}

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Saves the matrix to a given directx matrix. The extra row and column are set to the 
				///				same values as the identity.
				/// @param		m - matrix to save to
				//=====================================================================================================				

				inline void ToDxMatrix( D3DXMATRIX & m ) const
				{
					// Copy it:

					for ( int r = 0; r < 3 ; r++ )
					for ( int c = 0; c < 3 ; c++ )
					{
						m(r,c) = m_components[r][c];
					}	

					// Set the the extra row and column to the identity:

					m(3,0) = 0;
					m(3,1) = 0;
					m(3,2) = 0;
					m(3,3) = 1;

					m(0,3) = 0;
					m(1,3) = 0;
					m(2,3) = 0;
				}

				#endif

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Saves the matrix to a given directx matrix. The extra row and column are set to the 
				///				same values as the identity.
				/// @returns	A directx matrix
				//=====================================================================================================				

				inline D3DXMATRIX ToDxMatrix() const
				{
					// Make dx matrix:

					D3DXMATRIX m;

					// Copy it:

					for ( int r = 0; r < 3 ; r++ )
					for ( int c = 0; c < 3 ; c++ )
					{
						m(r,c) = m_components[r][c];
					}	

					// Set the the extra row and column to the identity:

					m(3,0) = 0;
					m(3,1) = 0;
					m(3,2) = 0;
					m(3,3) = 1;

					m(0,3) = 0;
					m(1,3) = 0;
					m(2,3) = 0;

					// Return 

					return m;
				}

				#endif

				//=====================================================================================================
				/// @brief		Gives the identity matrix
				/// @returns	The identity matrix
				//=====================================================================================================

				static inline Matrix3<T> Identity(){ return Matrix3<T>(); }

				//=====================================================================================================
				/// @brief		Transposes the matrix and returns the answer
				/// @returns	A new matrix which is the transpose of this matrix
				//=====================================================================================================

				static inline Matrix3<T> Transpose()
				{
					// Return the new matrix:

					return Matrix3<T>
					(
						m_components[0][0],m_components[1][0],m_components[2][0],
						m_components[0][1],m_components[1][1],m_components[2][1],
						m_components[0][2],m_components[1][2],m_components[2][2]
					);
				}

				//=====================================================================================================
				/// @brief		Creates a translation matrix to translate in 2d
				/// @param		v - vector containing the translation
				/// @returns	The given transform
				//=====================================================================================================

				template <class VT> static inline Matrix3<T> Translation( const Vector2<VT> & v )
				{
					// Return the matrix:

					return Matrix3<VT>
					(
						(VT) 1	, (VT) 0	, (VT) 0 ,
						(VT) 0	, (VT) 1	, (VT) 0 ,
						(VT) v.x, (VT) v.y	, (VT) 1 ,
					);
				}

				//=====================================================================================================
				/// @brief		Creates a translation matrix to translate in 2d
				/// @param		x - x translation amount
				/// @param		y - y translation amount
				/// @returns	The given transform
				//=====================================================================================================

				template <class VT> static inline Matrix3<T> Translation( VT x , VT y )
				{
					// Return the matrix:

					return Matrix3<T>
					(
						(T) 1	, (T) 0	, (T) 0 ,
						(T) 0	, (T) 1	, (T) 0 ,
						(T) x	, (T) y	, (T) 1 ,
					);
				}

				//=====================================================================================================
				/// @brief		Creates a rotation matrix to rotate about the x axis
				/// @param		r - amount of rotation about the x axis
				/// @returns	The given transform
				//=====================================================================================================

				template <class VT> static inline Matrix3<T> RotationX( VT r )
				{
					// Get cos and sin of the angle:

					float c = Float::Cos(r);
					float s = Float::sin(r);

					// Return the matrix:

					return Matrix3<T>
					(
						(T) 1	, (T) 0	 , (T) 0	,
						(T) 0	, (T) c  , (T) s	,
						(T) 0	, (T) -s , (T) c
					);
				}

				//=====================================================================================================
				/// @brief		Creates a rotation matrix to rotate about the y axis
				/// @param		r - amount of rotation about the y axis
				/// @returns	The given transform
				//=====================================================================================================

				template <class VT> static inline Matrix3<T> RotationY( VT r )
				{
					// Get cos and sin of the angle:

					float c = Float::Cos(r);
					float s = Float::Sin(r);

					// Return the matrix:

					return Matrix3<T>
					(
						(T) c	, (T) 0	, (T) -s ,
						(T) 0	, (T) 1 , (T) 0  ,
						(T) s	, (T) 0 , (T) c
					);
				}

				//=====================================================================================================
				/// @brief		Creates a rotation matrix to rotate about the z axis
				/// @param		r - amount of rotation about the z axis
				/// @returns	The given transform
				//=====================================================================================================

				template <class VT> static inline Matrix3<T> RotationZ( VT r )
				{
					// Get cos and sin of the angle:

					float c = Float::Cos(r);
					float s = Float::Sin(r);

					// Return the matrix:

					return Matrix3<T>
					(
						(T) c	, (T) s	, (T) 0 ,
						(T) -s	, (T) c , (T) 0 ,
						(T) 0	, (T) 0 , (T) 1
					);
				}

				//=====================================================================================================
				/// @brief		Creates a 2d rotation, which is equivalent to rotating about the z axis
				/// @param		r - amount of rotation
				/// @returns	The given transform
				//=====================================================================================================

				template <class VT> static inline Matrix3<T> Rotation( VT r ){ return RotationZ(r); }

				//=====================================================================================================
				/// @brief		Creates a scaling transform, in either 1d, 2d or 3d.
				/// @param		x - amount to scale on x axis
				/// @param		y - amount to scale on y axis
				/// @param		z - amount to scale on z axis
				/// @returns	The given scaling transform
				//=====================================================================================================

				template <class ST> static inline Matrix3<T> Scale( ST x , ST y = 1 , ST z = 1 )
				{ 
					return Matrix3<T>
					(
						(T) x , (T) 0 , (T) 0 ,
						(T) 0 , (T) y , (T) 0 ,
						(T) 0 , (T) 0 , (T) z 
					);
				}

				//=====================================================================================================
				/// @brief		Transforms the given vector and returns the answer.
				/// @param		v - vector to transform
				/// @returns	The vector transformed by the matrix
				/// @remarks	The w component remains unmodified. This function treats the vector as if it were 3d.
				//=====================================================================================================

				template <class VT> inline Vector4<VT> Transform( const Vector4<VT> & v ) const
				{ 
					// Save the matrix here:

					return Vector4<VT>
					(
						(VT)( v.x * m_components[0][0] + v.y * m_components[1][0] + v.z * m_components[2][0] ),
						(VT)( v.x * m_components[0][1] + v.y * m_components[1][1] + v.z * m_components[2][1] ),
						(VT)( v.x * m_components[0][2] + v.y * m_components[1][2] + v.z * m_components[2][2] ),
						v.w
					);
				}

				//=====================================================================================================
				/// @brief		Transforms the given vector and returns the answer.
				/// @param		v - vector to transform
				/// @returns	The vector transformed by the matrix
				//=====================================================================================================

				template <class VT> inline Vector3<VT> Transform( const Vector3<VT> & v ) const
				{ 
					// Save the matrix here:

					return Vector3<VT>
					(
						(VT)( v.x * m_components[0][0] + v.y * m_components[1][0] + v.z * m_components[2][0] ),
						(VT)( v.x * m_components[0][1] + v.y * m_components[1][1] + v.z * m_components[2][1] ),
						(VT)( v.x * m_components[0][2] + v.y * m_components[1][2] + v.z * m_components[2][2] )
					);
				}

				//=====================================================================================================
				/// @brief		Transforms the given vector and returns the answer. Treats vector as if
				///				z component is set to zero.
				/// @param		v - vector to transform
				/// @returns	The vector transformed by the matrix
				//=====================================================================================================

				template <class VT> inline Vector2<VT> Transform( const Vector2<VT> & v ) const
				{ 
					// Save the matrix here:

					return Vector2<VT>
					(
						(VT)( v.x * m_components[0][0] + v.y * m_components[1][0] ) ,
						(VT)( v.x * m_components[0][1] + v.y * m_components[1][1] )
					);
				}
		
			//=========================================================================================================
			// Variables
			//=========================================================================================================
			
				/// The components of the matrix

				T m_components[3][3];
	};
};

#endif 
