/*
	Trig Wars.
	Copyright (c) 2008, Darragh Coy.
	All rights reserved.
	Original work(s), source code & author(s) details available at: http://www.darraghcoy.com

	Redistribution and use in source and binary forms, with or without modification, are permitted 
	provided that the following conditions are met:

		* 	Redistributions of source code must retain the above copyright notice, this list of 
			conditions and the following disclaimer.
		* 	Redistributions in binary form must reproduce the above copyright notice, this list of 
			conditions and the following disclaimer in the documentation.
		* 	Neither the name of the author nor the names of contributors may be used to endorse or 
			promote products derived from this software without specific prior written permission.
		* 	The use of any third party libraries, modules or data files provided with the source 
			form of the program must follow and respect the individual licensing terms of these 
			components. You do not have permission to modify or circumvent the licensing terms for 
			provided third party components unless the third party licenses specifically allow this. 
			If you do not agree to the terms of provided third party components then you do not have 
			permission to use them in your derived work(s).
		* 	The terms of this license do not apply to separate (but related) distributions if supplied 
			by the author(s), only to the contents of this source distribution, derived works, and the 
			binaries produced by either modified or unmodified versions of the source code. The terms 
			of this license do not also apply to any third party libraries or components provided with 
			the source distribution. Such third party elements are licensed under their own separate 
			agreements.

	THIS SOFTWARE IS PROVIDED BY DARRAGH COY ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
	BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL DARRAGH COY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
	IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef GAME_MATRIX4_H
#define GAME_MATRIX4_H

//=====================================================================================================================
//=====================================================================================================================

#include "Base\Assertion.h"
#include "Base\Math\Float.h"
#include "Base\Math\Matrix2.h"
#include "Base\Math\Matrix3.h"
#include "Base\Math\Vector2.h"
#include "Base\Math\Vector3.h"
#include "Base\Math\Vector4.h"

// Specific:

#ifdef GAME_SUPPORT_DIRECT_X_TYPES

	#include <D3DX10Math.h>

#endif

//=====================================================================================================================
//=====================================================================================================================

namespace Game
{
	//=================================================================================================================
	/// @brief		Represents a 4x4 matrix
	//=================================================================================================================

	template <class T> struct Matrix4
	{
		public:

			//=========================================================================================================
			// Functions
			//=========================================================================================================

				//=====================================================================================================
				/// @brief		Constructor. Creates the identity matrix.
				//=====================================================================================================
				
				inline Matrix4()
				{
					m_components[0][0]=(T)1; m_components[0][1]=(T)0; m_components[0][2]=(T)0; m_components[0][3]=(T)0;
					m_components[1][0]=(T)0; m_components[1][1]=(T)1; m_components[1][2]=(T)0; m_components[1][3]=(T)0;
					m_components[2][0]=(T)0; m_components[2][1]=(T)0; m_components[2][2]=(T)1; m_components[2][3]=(T)0;
					m_components[3][0]=(T)0; m_components[3][1]=(T)0; m_components[3][2]=(T)0; m_components[3][3]=(T)1;
				}

				//=====================================================================================================
				/// @brief		Constructor. Creates from a 2x2 matrix and assigns missing components to the identity
				/// @param		m - matrix to construct from
				//=====================================================================================================
				
				template<class MT> inline Matrix4( Matrix2<MT> & m )
				{
					m_components[0][0]=(T) m.m_components[0][0];
					m_components[0][1]=(T) m.m_components[0][1];
					m_components[0][2]=(T) 0;
					m_components[0][3]=(T) 0;

					m_components[1][0]=(T) m.m_components[1][0];
					m_components[1][1]=(T) m.m_components[1][1];
					m_components[1][2]=(T) 0;
					m_components[1][3]=(T) 0;

					m_components[2][0]=(T) 0;
					m_components[2][1]=(T) 0;
					m_components[2][2]=(T) 1;
					m_components[2][3]=(T) 0;

					m_components[3][0]=(T) 0;
					m_components[3][1]=(T) 0;
					m_components[3][2]=(T) 0;
					m_components[3][3]=(T) 1;
				}

				//=====================================================================================================
				/// @brief		Constructor. Creates from a 3x3 matrix and assigns missing components to the identity
				/// @param		m - matrix to construct from
				//=====================================================================================================
				
				template<class MT> inline Matrix4( Matrix3<MT> & m )
				{
					m_components[0][0]=(T) m.m_components[0][0];
					m_components[0][1]=(T) m.m_components[0][1];
					m_components[0][2]=(T) m.m_components[0][2];
					m_components[0][3]=(T) 0;

					m_components[1][0]=(T) m.m_components[1][0];
					m_components[1][1]=(T) m.m_components[1][1];
					m_components[1][2]=(T) m.m_components[1][2];
					m_components[1][3]=(T) 0;

					m_components[2][0]=(T) m.m_components[2][0];
					m_components[2][1]=(T) m.m_components[2][1];
					m_components[2][2]=(T) m.m_components[2][2];
					m_components[2][3]=(T) 0;

					m_components[3][0]=(T) 0;
					m_components[3][1]=(T) 0;
					m_components[3][2]=(T) 0;
					m_components[3][3]=(T) 1;
				}

				//=====================================================================================================
				/// @brief		Constructor. Creates the matrix from the given elements.
				///	@param		r1c1 - 1st row 1st column
				///	@param		r1c2 - 1st row 2nd column
				///	@param		r1c3 - 1st row 3rd column
				///	@param		r1c4 - 1st row 4th column
				///	@param		r2c1 - 2nd row 1st column
				///	@param		r2c2 - 2nd row 2nd column
				///	@param		r2c3 - 2nd row 3rd column
				///	@param		r2c4 - 2nd row 4th column
				///	@param		r3c1 - 3rd row 1st column
				///	@param		r3c2 - 3rd row 3rd column
				///	@param		r3c3 - 3rd row 3rd column
				///	@param		r3c4 - 3rd row 4th column
				///	@param		r4c1 - 4th row 1st column
				///	@param		r4c2 - 4th row 2nd column
				///	@param		r4c3 - 4th row 3rd column
				///	@param		r4c4 - 4th row 4th column
				//=====================================================================================================
				
				template <class ET> inline Matrix4
				(
					ET r1c1,
					ET r1c2,
					ET r1c3,
					ET r1c4,
					ET r2c1,
					ET r2c2,
					ET r2c3,
					ET r2c4,
					ET r3c1,
					ET r3c2,
					ET r3c3,
					ET r3c4,
					ET r4c1,
					ET r4c2,
					ET r4c3,
					ET r4c4
				)
				{
					this->m_components[0][0] = (T) r1c1;
					this->m_components[0][1] = (T) r1c2;
					this->m_components[0][2] = (T) r1c3;
					this->m_components[0][3] = (T) r1c4;

					this->m_components[1][0] = (T) r2c1;
					this->m_components[1][1] = (T) r2c2;
					this->m_components[1][2] = (T) r2c3;
					this->m_components[1][3] = (T) r2c4;

					this->m_components[2][0] = (T) r3c1;
					this->m_components[2][1] = (T) r3c2;
					this->m_components[2][2] = (T) r3c3;
					this->m_components[2][3] = (T) r3c4;

					this->m_components[3][0] = (T) r4c1;
					this->m_components[3][1] = (T) r4c2;
					this->m_components[3][2] = (T) r4c3;
					this->m_components[3][3] = (T) r4c4;
				}

				//=====================================================================================================
				/// @brief		Constructor. Creates the matrix from a given array.
				///				Array must be row major. I.E the 1st row is stored 
				///				first in the array, second row next and so forth.
				///	@param		a - array to create from
				/// @pre		Given array cannot be null
				//=====================================================================================================
				
				template <class VT> inline Matrix4( VT * a )
				{
					// Make sure ok:

					ENSURE(a,"Given array cannot be null");

					// Do the copy:

					for ( int i = 0 ; i < 16 ; ++i )
					{
						*((T*)(m_components) + i ) = (T)(*( a + i ));
					}					
				}

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Constructor. Creates the matrix from a given DirectX 4x4 matrix, discarding the 
				///				translational part of the matrix.
				///	@param		m - matrix to create from
				//=====================================================================================================				

				inline Matrix4( const D3DXMATRIX & m )
				{
					// Copy it:

					for ( int r = 0; r < 4 ; r++ )
					for ( int c = 0; c < 4 ; c++ )
					{
						m_components[r][c] = m(r,c);
					}
				}

				#endif

				//=====================================================================================================
				/// @brief		Constructor. Creates the matrix from 4 column vectors.
				///	@param		a - array of column vectors
				/// @pre		Given array cannot be null
				//=====================================================================================================
				
				template <class VT> inline Matrix4( Vector4<VT> * a )
				{
					// Make sure ok:

					ENSURE(a,"Given array cannot be null");

					// Assign:

					for ( int i = 0 ; i < 4 ; ++i )
					{
						m_components[0][i] = (T) a[i].x;
						m_components[1][i] = (T) a[i].y;
						m_components[2][i] = (T) a[i].z;
						m_components[3][i] = (T) a[i].w;
					}
				}

				//=====================================================================================================
				/// @brief		Constructor. Creates the matrix from 4 column vectors.
				///	@param		c1 - column vector 1
				///	@param		c2 - column vector 2
				///	@param		c3 - column vector 3
				///	@param		c4 - column vector 4
				//=====================================================================================================
				
				template <class VT> inline Matrix4
				( 
					const Vector4<VT> & c1,
					const Vector4<VT> & c2,
					const Vector4<VT> & c3,
					const Vector4<VT> & c4
				)
				{
					m_components[0][0] = (T) c1.x;
					m_components[0][1] = (T) c2.x;
					m_components[0][2] = (T) c3.x;
					m_components[0][3] = (T) c4.x;

					m_components[1][0] = (T) c1.y;
					m_components[1][1] = (T) c2.y;
					m_components[1][2] = (T) c3.y;
					m_components[1][3] = (T) c4.y;

					m_components[2][0] = (T) c1.z;
					m_components[2][1] = (T) c2.z;
					m_components[2][2] = (T) c3.z;
					m_components[2][3] = (T) c4.z;

					m_components[3][0] = (T) c1.w;
					m_components[3][1] = (T) c2.w;
					m_components[3][2] = (T) c3.w;
					m_components[3][3] = (T) c4.w;
				}

				//=====================================================================================================
				/// @brief		Copy constructor. Makes this matrix from another
				/// @param		other - other matrix to assign from
				//=====================================================================================================

				template <class MT> inline Matrix4( const Matrix4<MT> & other )
				{
					// Do the copy:

					for ( int i = 0 ; i < 16 ; ++i )
					{
						*((T*)(m_components) + i ) = (T)(*((MT*)(other.m_components) + i));
					}
				}

				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the other matrix to this matrix. Makes up for the missing 
				///				components by setting them to the identity.
				/// @param		other - other matrix to assign from
				/// @returns	This matrix
				//=====================================================================================================
				
				template <class MT> inline Matrix4<T> & operator = ( const Matrix2<MT> & other )
				{
					m_components[0][0]=(T) other.m_components[0][0];
					m_components[0][1]=(T) other.m_components[0][1];
					m_components[0][2]=(T) 0;
					m_components[0][3]=(T) 0;

					m_components[1][0]=(T) other.m_components[1][0];
					m_components[1][1]=(T) other.m_components[1][1];
					m_components[1][2]=(T) 0;
					m_components[1][3]=(T) 0;

					m_components[2][0]=(T) 0;
					m_components[2][1]=(T) 0;
					m_components[2][2]=(T) 1;
					m_components[2][3]=(T) 0;

					m_components[3][0]=(T) 0;
					m_components[3][1]=(T) 0;
					m_components[3][2]=(T) 0;
					m_components[3][3]=(T) 1;

					return * this;
				}

				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the other matrix to this matrix. Makes up for the missing 
				///				components by setting them to the identity.
				/// @param		other - other matrix to assign from
				/// @returns	This matrix
				//=====================================================================================================
				
				template <class MT> inline Matrix4<T> & operator = ( const Matrix3<MT> & other )
				{
					m_components[0][0]=(T) other.m_components[0][0];
					m_components[0][1]=(T) other.m_components[0][1];
					m_components[0][2]=(T) other.m_components[0][2];
					m_components[0][3]=(T) 0;

					m_components[1][0]=(T) other.m_components[1][0];
					m_components[1][1]=(T) other.m_components[1][1];
					m_components[1][2]=(T) other.m_components[1][2];
					m_components[1][3]=(T) 0;

					m_components[2][0]=(T) other.m_components[2][0];
					m_components[2][1]=(T) other.m_components[2][1];
					m_components[2][2]=(T) other.m_components[2][2];
					m_components[2][3]=(T) 0;

					m_components[3][0]=(T) 0;
					m_components[3][1]=(T) 0;
					m_components[3][2]=(T) 0;
					m_components[3][3]=(T) 1;

					return * this;
				}

				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the other matrix to this matrix.
				/// @param		other - other matrix to assign from
				/// @returns	This matrix
				//=====================================================================================================

				template <class MT> inline Matrix4<T> & operator = ( const Matrix4<MT> & other )
				{
					// Do the copy:

					for ( int i = 0 ; i < 16 ; ++i )
					{
						*((T*)(m_components) + i ) = (T)(*((MT*)(other.m_components) + i));
					}

					return * this;
				}

				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the given array of column vectors to the matrix
				///	@param		a - array representing a matrix
				/// @returns	This matrix
				/// @pre		Given array cannot be null	
				//=====================================================================================================
				
				template <class VT> inline Matrix4<T> & operator = ( const Vector4<VT> * a )
				{
					// Make sure ok:

					ENSURE(a,"Given array cannot be null");

					// Assign:

					for ( int i = 0 ; i < 4 ; ++i )
					{
						m_components[0][i] = (T) a[i].x;
						m_components[1][i] = (T) a[i].y;
						m_components[2][i] = (T) a[i].z;
						m_components[3][i] = (T) a[i].w;
					}

					return * this;
				}

				//=====================================================================================================
				/// @brief		Assignment operator. Creates the matrix from a given array.
				///				Array must be row major. I.E the 1st row is stored 
				///				first in the array, second row next and so forth.
				///	@param		a - array to create from
				/// @returns	This matrix
				/// @pre		Given array cannot be null				
				//=====================================================================================================
				
				template <class VT> inline Matrix4<T> & operator = ( VT * a )
				{
					// Make sure ok:

					ENSURE(a,"Given array cannot be null");

					// Do the copy:

					for ( int i = 0 ; i < 16 ; ++i )
					{
						*((T*)(m_components) + i ) = (T)(*( a + i ));
					}		

					return * this;
				}

				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the given 4x4 directx matrix to this matrix.
				///	@param		m - matrix to assign to this matrix
				/// @returns	this
				//=====================================================================================================

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES

					inline Matrix3<T> & operator = ( const D3DXMATRIX & m )
					{
						// Copy it:

						for ( int r = 0; r < 4 ; r++ )
						for ( int c = 0; c < 4 ; c++ )
						{
							m_components[r][c] = m(r,c);
						}

						return * this;
					}

				#endif

				//=====================================================================================================
				/// @brief		Matrix multiplication. Multiplies this matrix with another matrix.
				/// @param		m - matrix to multiply with
				/// @returns	a new matrix
				//=====================================================================================================

				template <class MT> inline Matrix4<T> operator * ( const Matrix4<MT> & m ) const
				{
					// Need to make a temporary copy of this matrix:

					Matrix4<T> new_matrix;

					// Do this to shorten things:

					T  * a = (T* ) this->m_components;					
					MT * b = (MT*) m.m_components;
					T  * c = (T* ) new_matrix.m_components;
					
					// Right: on to the multiply

					c[0 ] = (T)( a[0 ] * b[0 ] + a[1 ] * b[4 ] + a[2 ] * b[8 ] + a[3 ] * b[12] ); 
					c[1 ] = (T)( a[0 ] * b[1 ] + a[1 ] * b[5 ] + a[2 ] * b[9 ] + a[3 ] * b[13] );
					c[2 ] = (T)( a[0 ] * b[2 ] + a[1 ] * b[6 ] + a[2 ] * b[10] + a[3 ] * b[14] );
					c[3 ] = (T)( a[0 ] * b[3 ] + a[1 ] * b[7 ] + a[2 ] * b[11] + a[3 ] * b[15] );

					c[4 ] = (T)( a[4 ] * b[0 ] + a[5 ] * b[4 ] + a[6 ] * b[8 ] + a[7 ] * b[12] ); 
					c[5 ] = (T)( a[4 ] * b[1 ] + a[5 ] * b[5 ] + a[6 ] * b[9 ] + a[7 ] * b[13] );
					c[6 ] = (T)( a[4 ] * b[2 ] + a[5 ] * b[6 ] + a[6 ] * b[10] + a[7 ] * b[14] );
					c[7 ] = (T)( a[4 ] * b[3 ] + a[5 ] * b[7 ] + a[6 ] * b[11] + a[7 ] * b[15] );

					c[8 ] = (T)( a[8 ] * b[0 ] + a[9 ] * b[4 ] + a[10] * b[8 ] + a[11] * b[12] ); 
					c[9 ] = (T)( a[8 ] * b[1 ] + a[9 ] * b[5 ] + a[10] * b[9 ] + a[11] * b[13] );
					c[10] = (T)( a[8 ] * b[2 ] + a[9 ] * b[6 ] + a[10] * b[10] + a[11] * b[14] );
					c[11] = (T)( a[8 ] * b[3 ] + a[9 ] * b[7 ] + a[10] * b[11] + a[11] * b[15] );

					c[12] = (T)( a[12] * b[0 ] + a[13] * b[4 ] + a[14] * b[8 ] + a[15] * b[12] ); 
					c[13] = (T)( a[12] * b[1 ] + a[13] * b[5 ] + a[14] * b[9 ] + a[15] * b[13] );
					c[14] = (T)( a[12] * b[2 ] + a[13] * b[6 ] + a[14] * b[10] + a[15] * b[14] );
					c[15] = (T)( a[12] * b[3 ] + a[13] * b[7 ] + a[14] * b[11] + a[15] * b[15] );

					// Return the new matrix

					return new_matrix;
				}

				//=====================================================================================================
				/// @brief		Matrix multiplication. Multiplies this matrix with another matrix.
				/// @param		m - matrix to multiply with
				/// @returns	this matrix
				//=====================================================================================================

				template <class MT> inline void operator *= ( const Matrix4<MT> & m )
				{
					// Need to make a temporary copy of this matrix:

					Matrix4<T> temp(*this);

					// Do this to shorten things:

					T  * a = (T* ) temp.m_components;
					MT * b = (MT*) m.m_components
					T  * c = (T* ) m_components;
					
					// Right: on to the multiply

					c[0 ] = (T)( a[0 ] * b[0 ] + a[1 ] * b[4 ] + a[2 ] * b[8 ] + a[3 ] * b[12] ); 
					c[1 ] = (T)( a[0 ] * b[1 ] + a[1 ] * b[5 ] + a[2 ] * b[9 ] + a[3 ] * b[13] );
					c[2 ] = (T)( a[0 ] * b[2 ] + a[1 ] * b[6 ] + a[2 ] * b[10] + a[3 ] * b[14] );
					c[3 ] = (T)( a[0 ] * b[3 ] + a[1 ] * b[7 ] + a[2 ] * b[11] + a[3 ] * b[15] );

					c[4 ] = (T)( a[4 ] * b[0 ] + a[5 ] * b[4 ] + a[6 ] * b[8 ] + a[7 ] * b[12] ); 
					c[5 ] = (T)( a[4 ] * b[1 ] + a[5 ] * b[5 ] + a[6 ] * b[9 ] + a[7 ] * b[13] );
					c[6 ] = (T)( a[4 ] * b[2 ] + a[5 ] * b[6 ] + a[6 ] * b[10] + a[7 ] * b[14] );
					c[7 ] = (T)( a[4 ] * b[3 ] + a[5 ] * b[7 ] + a[6 ] * b[11] + a[7 ] * b[15] );

					c[8 ] = (T)( a[8 ] * b[0 ] + a[9 ] * b[4 ] + a[10] * b[8 ] + a[11] * b[12] ); 
					c[9 ] = (T)( a[8 ] * b[1 ] + a[9 ] * b[5 ] + a[10] * b[9 ] + a[11] * b[13] );
					c[10] = (T)( a[8 ] * b[2 ] + a[9 ] * b[6 ] + a[10] * b[10] + a[11] * b[14] );
					c[11] = (T)( a[8 ] * b[3 ] + a[9 ] * b[7 ] + a[10] * b[11] + a[11] * b[15] );

					c[12] = (T)( a[12] * b[0 ] + a[13] * b[4 ] + a[14] * b[8 ] + a[15] * b[12] ); 
					c[13] = (T)( a[12] * b[1 ] + a[13] * b[5 ] + a[14] * b[9 ] + a[15] * b[13] );
					c[14] = (T)( a[12] * b[2 ] + a[13] * b[6 ] + a[14] * b[10] + a[15] * b[14] );
					c[15] = (T)( a[12] * b[3 ] + a[13] * b[7 ] + a[14] * b[11] + a[15] * b[15] );
				}

				//=====================================================================================================
				/// @brief		Retrieves the given row vector.
				/// @param		r - row to retrieve.
				/// @pre		Row index must be in range
				//======================================================================================================

				inline Vector4<T> GetRow( unsigned r ) const
				{
					ENSURE( r < 4 , "Row index must be in range" );

					return Vector4<T>
					(
						m_components[r][0],
						m_components[r][1],
						m_components[r][2],
						m_components[r][3]
					);
				}

				//=====================================================================================================
				/// @brief		Sets the given row to the given value
				/// @param		r - row to set
				/// @param		v - vector representing the row
				/// @pre		Row index must be in range
				//======================================================================================================

				template<class VT> inline void SetRow( unsigned r , const Vector4<VT> & v )
				{
					ENSURE( r < 4 , "Row index must be in range" );

					m_components[r][0] = (T) v.x;
					m_components[r][1] = (T) v.y;
					m_components[r][2] = (T) v.z;
					m_components[r][3] = (T) v.w;
				}

				//=====================================================================================================
				/// @brief		Retrieves the given column vector.
				/// @param		c - column to retrieve.
				/// @pre		Column index must be in range
				//======================================================================================================

				inline Vector4<T> GetCol( unsigned c ) const
				{
					ENSURE( c < 4 , "Column index must be in range" );

					return Vector4<T>
					(
						m_components[0][c],
						m_components[1][c],
						m_components[2][c],
						m_components[3][c]
					);
				}

				//=====================================================================================================
				/// @brief		Sets the given column to the given value
				/// @param		c - column to set
				/// @param		v - vector representing the row
				/// @pre		Column index must be in range
				//======================================================================================================

				template<class VT> inline void SetCol( unsigned c , const Vector4<VT> & v )
				{
					ENSURE( c < 4 , "Column index must be in range" );

					m_components[0][c] = (T) v.x;
					m_components[1][c] = (T) v.y;
					m_components[2][c] = (T) v.z;
					m_components[3][c] = (T) v.w;
				}

				//=====================================================================================================
				/// @brief		Saves the matrix to a given array. Matrix is saved in row major format.
				/// @param		a - array to save to.
				/// @pre		Given array cannot be null
				//=====================================================================================================

				template <class AT> inline void ToArray( AT * a ) const
				{
					// Save:

					for ( int i = 0 ; i < 16 ; ++i )
					{
						* (a + i) = (AT) * (((T*)(m_components)) + i );
					}
				}

				//=====================================================================================================
				/// @brief		Converts to a 2x2 matrix by discarding the extra 4x4 bits
				/// @returns	This matrix in it's 2x2 form
				//=====================================================================================================

				inline void ToMatrix2() const
				{
					return Matrix2<T>
					(
						m_components[0][0],
						m_components[0][1],
						m_components[1][0],
						m_components[1][1]
					);
				}

				//=====================================================================================================
				/// @brief		Converts to a 3x3 matrix by discarding the extra 4x4 bits
				/// @returns	This matrix in it's 3x3 form
				//=====================================================================================================

				inline void ToMatrix3() const
				{
					return Matrix3<T>
					(
						m_components[0][0],
						m_components[0][1],
						m_components[0][2],
						m_components[1][0],
						m_components[1][1],
						m_components[1][2],
						m_components[2][0],
						m_components[2][1],
						m_components[2][2]
					);
				}

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Saves the matrix to a given directx matrix.
				///				same values as the identity.
				/// @param		m - matrix to save to
				//=====================================================================================================				

				inline void ToDxMatrix( D3DXMATRIX & m ) const
				{
					// Copy it:

					for ( int r = 0; r < 4 ; r++ )
					for ( int c = 0; c < 4 ; c++ )
					{
						m(r,c) = m_components[r][c];
					}
				}

				#endif

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Saves the matrix to a given directx matrix.
				///				same values as the identity.
				/// @returns	A directx matrix
				//=====================================================================================================				

				inline D3DXMATRIX ToDxMatrix() const
				{
					// Make a dx matrix

					D3DXMATRIX m;

					// Copy it:

					for ( int r = 0; r < 4 ; r++ )
					for ( int c = 0; c < 4 ; c++ )
					{
						m(r,c) = m_components[r][c];
					}

					// Return it:

					return m;
				}
					
				#endif

				//=====================================================================================================
				/// @brief		Gives the identity matrix
				/// @returns	The identity matrix
				//=====================================================================================================

				static inline Matrix4<T> Identity()
				{ 
					return Matrix4<T>(); 
				}

				//=====================================================================================================
				/// @brief		Transposes the matrix and returns the answer
				/// @returns	A new matrix which is the transpose of this matrix
				//=====================================================================================================

				inline Matrix4<T> Transpose()
				{
					// Return the new matrix:

					return Matrix4<T>
					(
						m_components[0][0],m_components[1][0],m_components[2][0],m_components[3][0],
						m_components[0][1],m_components[1][1],m_components[2][1],m_components[3][1],
						m_components[0][2],m_components[1][2],m_components[2][2],m_components[3][2],
						m_components[0][3],m_components[1][3],m_components[2][3],m_components[3][3]
					);
				}

				//=====================================================================================================
				/// @brief		Creates a translation matrix to translate in 2d
				/// @param		v - vector containing the translation
				/// @returns	The given transform
				//=====================================================================================================

				template <class VT> static inline Matrix4<VT> Translation( const Vector2<VT> & v )
				{
					// Return the matrix:

					return Matrix4<VT>
					(
						(VT) 1	, (VT) 0	, (VT) 0	, (VT) 0,
						(VT) 0	, (VT) 1	, (VT) 0	, (VT) 0,
						(VT) 0	, (VT) 0	, (VT) 1	, (VT) 0,
						(VT) v.x, (VT) v.y	, (VT) 0	, (VT) 1
					);
				}

				//=====================================================================================================
				/// @brief		Creates a translation matrix to translate in 3d
				/// @param		v - vector containing the translation
				/// @returns	The given transform
				//=====================================================================================================

				template <class VT> static inline Matrix4<VT> Translation( const Vector3<VT> & v )
				{
					// Return the matrix:

					return Matrix4<VT>
					(
						(VT) 1 , (VT) 0	, (VT) 0	, (VT) v.x ,
						(VT) 0 , (VT) 1	, (VT) 0	, (VT) v.y ,
						(VT) 0 , (VT) 0	, (VT) 1	, (VT) v.z ,
						(VT) 0 , (VT) 0	, (VT) 0	, (VT) 1
					);
				}

				//=====================================================================================================
				/// @brief		Creates a translation matrix to translate in 3d
				/// @param		v - vector containing the translation
				/// @returns	The given transform
				//=====================================================================================================

				template <class VT> static inline Matrix4<VT> Translation( const Vector4<VT> & v )
				{
					// Return the matrix:

					return Matrix4<VT>
					(
						(VT) 1	, (VT) 0		, (VT) 0		, (VT) 0,
						(VT) 0	, (VT) 1		, (VT) 0		, (VT) 0,
						(VT) 0	, (VT) 0		, (VT) 1		, (VT) 0,
						(VT) v.x, (VT) v.y		, (VT) v.z		, (VT) 1
					);
				}

				//=====================================================================================================
				/// @brief		Creates a translation matrix to translate in 3d
				/// @param		x - x translation amount
				/// @param		y - y translation amount
				/// @param		z - z translation amount
				/// @returns	The given transform
				//=====================================================================================================

				template <class VT> static inline Matrix4<VT> Translation( VT x , VT y , VT z )
				{
					// Return the matrix:

					return Matrix4<VT>
					(
						(VT) 1	, (VT) 0	, (VT) 0 , (VT) 0 ,
						(VT) 0	, (VT) 1	, (VT) 0 , (VT) 0 ,
						(VT) 0	, (VT) 0	, (VT) 1 , (VT) 0 ,
						(VT) x	, (VT) y	, (VT) z , (VT) 1
					);
				}

				//=====================================================================================================
				/// @brief		Creates a rotation matrix to rotate about the x axis
				/// @param		r - amount of rotation about the x axis
				/// @returns	The given transform
				//=====================================================================================================

				template <class VT> static inline Matrix4<T> RotationX( VT r )
				{
					// Get cos and sin of the angle:

					float c = Float::Cos(r);
					float s = Float::Sin(r);

					// Return the matrix:

					return Matrix4<T>
					(
						(T) 1	, (T) 0	 , (T) 0	, (T) 0	,
						(T) 0	, (T) c  , (T) s	, (T) 0 ,
						(T) 0	, (T) -s , (T) c	, (T) 0 ,
						(T) 0	, (T) 0	 , (T) 0	, (T) 1
					);
				}

				//=====================================================================================================
				/// @brief		Creates a rotation matrix to rotate about the y axis
				/// @param		r - amount of rotation about the y axis
				/// @returns	The given transform
				//=====================================================================================================

				template <class VT> static inline Matrix4<T> RotationY( VT r )
				{
					// Get cos and sin of the angle:

					float c = Float::Cos(r);
					float s = Float::Sin(r);

					// Return the matrix:

					return Matrix4<T>
					(
						(T) c	, (T) 0	, (T) -s , (T) 0 ,
						(T) 0	, (T) 1 , (T) 0  , (T) 0 ,
						(T) s	, (T) 0 , (T) c	 , (T) 0 ,
						(T) 0	, (T) 0 , (T) 0  , (T) 1
					);
				}

				//=====================================================================================================
				/// @brief		Creates a rotation matrix to rotate about the z axis
				/// @param		r - amount of rotation about the z axis
				/// @returns	The given transform
				//=====================================================================================================

				template <class VT> static inline Matrix4<T> RotationZ( VT r )
				{
					// Get cos and sin of the angle:

					float c = Float::Cos(r);
					float s = Float::Sin(r);

					// Return the matrix:

					return Matrix4<T>
					(
						(T) c	, (T) s	, (T) 0 , (T) 0 , 
						(T) -s	, (T) c , (T) 0 , (T) 0	,
						(T) 0	, (T) 0 , (T) 1 , (T) 0 ,
						(T) 0	, (T) 0 , (T) 0 , (T) 1
					);
				}

				//=====================================================================================================
				/// @brief		Creates a scaling transform, in either 1d, 2d or 3d.
				/// @param		x - amount to scale on x axis
				/// @param		y - amount to scale on y axis
				/// @param		z - amount to scale on z axis
				/// @returns	The given scaling transform
				//=====================================================================================================

				template <class ST> static inline Matrix4<T> Scale( ST x , ST y = 1 , ST z = 1 )
				{ 
					return Matrix4<T>
					(
						(T) x , (T) 0 , (T) 0 , (T) 0 ,
						(T) 0 , (T) y , (T) 0 , (T) 0 ,
						(T) 0 , (T) 0 , (T) z , (T) 0 ,
						(T) 0 , (T) 0 , (T) 0 , (T) 1
					);
				}

				//=====================================================================================================
				/// @brief		Creates a scaling transform
				/// @param		s - amount to scale by
				/// @returns	The given scaling transform
				//=====================================================================================================

				template <class VT> static inline Matrix4<T> Scale( const Vector3<VT> & s )
				{ 
					return Matrix4<T>
					(
						(T) s.x , (T) 0		, (T) 0		, (T) 0 ,
						(T) 0	, (T) s.y	, (T) 0		, (T) 0 ,
						(T) 0	, (T) 0		, (T) s.z	, (T) 0 ,
						(T) 0	, (T) 0		, (T) 0		, (T) 1
					);
				}

				//=====================================================================================================
				/// @brief		Creates a right handed orthographic projection matrix
				/// @param		w - width of the view
				/// @param		h - height of the view
				/// @param		zn - near z plane value
				/// @param		zf - far z plane value
				/// @returns	Projection matrix with the given parameters
				//=====================================================================================================

				template <class ST> static inline Matrix4<T> Orthographic
				( 
					ST w,
					ST h,
					ST zn,
					ST zf
				)
				{ 
					// This is based on the D3DXMatrixOrthoRH() method documented in the DirectX SDK

					return Matrix4<T>
					(
						(T)(2.0f /w )	, (T) 0					, (T) 0						, (T) 0 ,
						(T) 0			, (T)( 2.0f / h )		, (T) 0						, (T) 0 ,
						(T) 0			, (T) 0					, (T)( 1.0f / ( zn - zf ) )	, (T) 0 ,
						(T) 0			, (T) 0					, (T)( zn / ( zn - zf ) )	, (T) 1 
					);
				}

				//=====================================================================================================
				/// @brief		Transforms the given vector and returns the answer.
				/// @param		v - vector to transform
				/// @returns	The vector transformed by the matrix
				//=====================================================================================================

				template <class VT> inline Vector4<VT> Transform( const Vector4<VT> & v ) const
				{ 
					// Save the matrix here:

					return Vector4<VT>
					(
						(VT)( v.x * m_components[0][0] + v.y * m_components[1][0] + v.z * m_components[2][0] + v.w * m_components[3][0] ),
						(VT)( v.x * m_components[0][1] + v.y * m_components[1][1] + v.z * m_components[2][1] + v.w * m_components[3][1] ),
						(VT)( v.x * m_components[0][2] + v.y * m_components[1][2] + v.z * m_components[2][2] + v.w * m_components[3][2] ),
						(VT)( v.x * m_components[0][3] + v.y * m_components[1][3] + v.z * m_components[2][3] + v.w * m_components[3][3] )
					);
				}

				//=====================================================================================================
				/// @brief		Transforms the given vector and returns the answer. Treats vector as if
				///				w component is set to 1.
				/// @param		v - vector to transform
				/// @returns	The vector transformed by the matrix
				//=====================================================================================================

				template <class VT> inline Vector3<VT> Transform( const Vector3<VT> & v ) const
				{ 
					// Save the matrix here:

					return Vector3<VT>
					(
						(VT)( v.x * m_components[0][0] + v.y * m_components[1][0] + v.z * m_components[2][0] + m_components[3][0] ),
						(VT)( v.x * m_components[0][1] + v.y * m_components[1][1] + v.z * m_components[2][1] + m_components[3][1] ),
						(VT)( v.x * m_components[0][2] + v.y * m_components[1][2] + v.z * m_components[2][2] + m_components[3][2] )
					);
				}

				//=====================================================================================================
				/// @brief		Transforms the given vector and returns the answer. Treats vector as if
				///				w component is set to 1, and z component is set to 0.
				/// @param		v - vector to transform
				/// @returns	The vector transformed by the matrix
				//=====================================================================================================

				template <class VT> inline Vector2<VT> Transform( const Vector2<VT> & v ) const
				{ 
					// Save the matrix here:

					return Vector2<VT>
					(
						(VT)( v.x * m_components[0][0] + v.y * m_components[1][0] + m_components[3][0] ),
						(VT)( v.x * m_components[0][1] + v.y * m_components[1][1] + m_components[3][1] )
					);
				}

			//=========================================================================================================
			// Variables
			//=========================================================================================================
			
				/// The components of the matrix

				T m_components[4][4];
	};
};

#endif 
