/*
	Trig Wars.
	Copyright (c) 2008, Darragh Coy.
	All rights reserved.
	Original work(s), source code & author(s) details available at: http://www.darraghcoy.com

	Redistribution and use in source and binary forms, with or without modification, are permitted 
	provided that the following conditions are met:

		* 	Redistributions of source code must retain the above copyright notice, this list of 
			conditions and the following disclaimer.
		* 	Redistributions in binary form must reproduce the above copyright notice, this list of 
			conditions and the following disclaimer in the documentation.
		* 	Neither the name of the author nor the names of contributors may be used to endorse or 
			promote products derived from this software without specific prior written permission.
		* 	The use of any third party libraries, modules or data files provided with the source 
			form of the program must follow and respect the individual licensing terms of these 
			components. You do not have permission to modify or circumvent the licensing terms for 
			provided third party components unless the third party licenses specifically allow this. 
			If you do not agree to the terms of provided third party components then you do not have 
			permission to use them in your derived work(s).
		* 	The terms of this license do not apply to separate (but related) distributions if supplied 
			by the author(s), only to the contents of this source distribution, derived works, and the 
			binaries produced by either modified or unmodified versions of the source code. The terms 
			of this license do not also apply to any third party libraries or components provided with 
			the source distribution. Such third party elements are licensed under their own separate 
			agreements.

	THIS SOFTWARE IS PROVIDED BY DARRAGH COY ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
	BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL DARRAGH COY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
	IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef GAME_VECTOR4_H
#define GAME_VECTOR4_H

//=====================================================================================================================
//=====================================================================================================================

#include "Base\Math\Vector2.h"
#include "Base\Math\Vector3.h"
#include "Base\Assertion.h"
#include <cmath>

// Specific:

#ifdef GAME_SUPPORT_DIRECT_X_TYPES

	#include <D3DX10Math.h>

#endif

//=====================================================================================================================
//=====================================================================================================================

namespace Game
{
	//=================================================================================================================
	/// @brief		Four dimensional vector object. Includes all the standard vector operations.
	//=================================================================================================================

	template <class T> struct Vector4
	{
		public:

			//=========================================================================================================
			// Functions
			//=========================================================================================================

				//=====================================================================================================
				/// @brief		Constructor. Creates the vector.
				/// @param		x - x component of the vector
				/// @param		y - y component of the vector
				/// @param		z - y component of the vector
				/// @param		w - w component of the vector
				//=====================================================================================================
				
				inline Vector4( T x = 0 , T y = 0 , T z = 0 , T w = (T) 1 )
				{
					this->x = (T) x;
					this->y = (T) y;
					this->z = (T) z;
					this->w = (T) w;
				}			
				
				//=====================================================================================================
				/// @brief		Constructor. Creates this vector from another 3d vector and a w component if given.
				/// @param		v - other vector to construct from
				/// @param		w - w component of the vector
				/// @remarks	If w coordinate is not given it is set to 1
				//=====================================================================================================
				
				template<class VT> inline Vector4( const Vector3<VT> & v , T w = (T) 1 )
				{
					this->x = (T) v.x;
					this->y = (T) v.y;
					this->z = (T) v.z;
					this->w = w;
				}
				
				//=====================================================================================================
				/// @brief		Constructor. Creates this vector from another 2d vector and a given z and w coordinate,
				///				if specified.
				/// @param		z - z component of the vector
				/// @param		w - w component of the vector
				/// @param		v - other vector to construct from
				/// @remarks	If z component is not given then it is set to 0
				/// @remarks	If w component is not given then it is set to 1
				//=====================================================================================================
				
				template<class VT> inline Vector4( const Vector2<VT> & v , T z = (T) 0 , T w = (T) 1 )
				{
					this->x = (T) v.x;
					this->y = (T) v.y;
					this->z = z;
					this->w = w;
				}										
				
				//=====================================================================================================
				/// @brief		Constructor. Creates the vector from an array
				/// @param		a - array to create from
				/// @pre		Given pointer must not be null
				//=====================================================================================================
				
				template<class VT> inline Vector4( const VT * a )
				{
					ENSURE(a,"Given pointer must not be null");
				
					this->x = (T) a[0];
					this->y = (T) a[1];
					this->z = (T) a[2];
					this->w = (T) a[3];
				}

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Constructor. Creates the vector from a given directx 2d vector and sets z to 0, and w 
				///				to 1
				/// @param		v - vector to create from
				//=====================================================================================================				
	
				inline Vector4( const D3DXVECTOR2 & v )
				{
					this->x = (T) v.x;
					this->y = (T) v.y;
					this->z = (T) 0;
					this->w = (T) 1;
				}					

				#endif

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Constructor. Creates the vector from a given directx 3d vector and sets w to 1
				/// @param		v - vector to create from
				//=====================================================================================================				
	
				inline Vector4( const D3DXVECTOR3 & v )
				{
					this->x = (T) v.x;
					this->y = (T) v.y;
					this->z = (T) v.z;
					this->w = (T) 1;
				}					

				#endif

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Constructor. Creates the vector from a given directx 4d vector
				/// @param		v - vector to create from
				//=====================================================================================================				
	
				inline Vector4( const D3DXVECTOR4 & v )
				{
					this->x = (T) v.x;
					this->y = (T) v.y;
					this->z = (T) v.z;
					this->w = (T) v.w;
				}					

				#endif
				
				//=====================================================================================================
				/// @brief		Copy constructor. Creates this vector from another vector.
				/// @param		v - other vector to construct from
				//=====================================================================================================
				
				template<class VT> inline Vector4( const Vector4<VT> & v )
				{
					this->x = (T) v.x;
					this->y = (T) v.y;
					this->z = (T) v.z;
					this->w = (T) v.w;
				}					
				
				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the other vector to this vector.
				/// @param		v - other vector to assign from
				/// @returns	This vector
				//=====================================================================================================
				
				template<class VT> inline Vector4<T> & operator = ( const Vector4<VT> & v )
				{
					this->x = (T) v.x;
					this->y = (T) v.y;
					this->z = (T) v.z;
					this->w = (T) v.w;
					
					return * this;
				}
				
				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the 3d vector to this vector and sets w to 1
				/// @param		v - other vector to assign from
				/// @returns	This vector
				//=====================================================================================================
				
				template<class VT> inline Vector4<T> & operator = ( const Vector3<VT> & v )
				{
					this->x = (T) v.x;
					this->y = (T) v.y;
					this->z = (T) v.z;
					this->w = 1;
					
					return * this;
				}
				
				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the 2d vector to this vector and sets z to 0, and w to 1
				/// @param		v - other vector to assign from
				/// @returns	This vector
				//=====================================================================================================
				
				template<class VT> inline Vector4<T> & operator = ( const Vector2<VT> & v )
				{
					this->x = (T) v.x;
					this->y = (T) v.y;
					this->z = 0;
					this->w = 1;
					
					return * this;
				}											
				
				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the given array to this vector,
				/// @param		a - array to assign to this vector.
				/// @returns	This vector
				//=====================================================================================================
				
				template<class VT> inline Vector3<T> & operator = ( const VT * a )
				{				
					this->x = (T) v[0];
					this->y = (T) v[1];
					this->z = (T) v[2];
					this->w = (T) v[3];
					
					return * this;
				}

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the given 2d directx vector to this vector and sets z to 0,
				///				and w to 1
				/// @param		v - vector to assign
				/// @returns	this vector
				//=====================================================================================================				
	
				inline Vector4<T> & operator = ( const D3DXVECTOR2 & v )
				{
					this->x = (T) v.x;
					this->y = (T) v.y;
					this->z = (T) 0;
					this->w = (T) 1;

					return * this;
				}					

				#endif

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the given 3d directx vector to this vector, and sets w
				///				to 1.
				/// @param		v - vector to assign
				/// @returns	this vector
				//=====================================================================================================				
	
				inline Vector4<T> & operator = ( const D3DXVECTOR3 & v )
				{
					this->x = (T) v.x;
					this->y = (T) v.y;
					this->z = (T) v.z;
					this->w = (T) 1;

					return * this;
				}					

				#endif

				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Assignment operator. Assigns the given 4d directx vector to this vector.
				/// @param		v - vector to assign
				/// @returns	this vector
				//=====================================================================================================				
	
				inline Vector4<T> & operator = ( const D3DXVECTOR4 & v )
				{
					this->x = (T) v.x;
					this->y = (T) v.y;
					this->z = (T) v.z;
					this->w = (T) v.w;

					return * this;
				}					

				#endif
				
				//=====================================================================================================
				/// @brief		Array index accessor. Returns a copy of the variable at the given index.
				/// @param		i - element to access
				/// @returns	chosen element
				/// @pre		Given index must be in range
				//=====================================================================================================
				
				inline T operator[]( unsigned i ) const
				{
					ENSURE(i<4,"Given index must be in range");
				
					return *((&x)+i);
				}
				
				//=====================================================================================================
				/// @brief		Array index accessor. Returns a reference of the variable at the given index.
				/// @param		i - element to access
				/// @returns	chosen element, ready for writing
				/// @pre		Given index must be in range
				//=====================================================================================================
				
				inline T & operator[]( unsigned i )
				{
					ENSURE(i<4,"Given index must be in range");
				
					return *((&x)+i);
				}																		
				
				//=====================================================================================================
				/// @brief		Addition. Adds the given vector to this vector.
				/// @param		v - other vector to add to this vector
				/// @returns	A new vector
				//=====================================================================================================
				
				template<class VT> inline Vector4<T> operator + ( const Vector4<VT> & v ) const
				{
					return Vector4<T>
					( 
						(T)( v.x + this->x ), 
						(T)( v.y + this->y ),
						(T)( v.z + this->z ),
						(T)( v.w + this->w )
					);
				}
				
				//=====================================================================================================
				/// @brief		Subtraction. Subtracts the given vector from this vector.
				/// @param		v - other vector to subtract from this vector
				/// @returns	A new vector
				//=====================================================================================================
				
				template<class VT> inline Vector4<T> operator - ( const Vector4<VT> & v ) const
				{
					return Vector4<T>
					( 
						(T)( this->x - v.x ), 
						(T)( this->y - v.y ),
						(T)( this->z - v.z ),
						(T)( this->w - v.w )
					);
				}							
				
				//=====================================================================================================
				/// @brief		Scalar product. Returns a new vector scaled by the given amount.
				/// @param		scale - scaling amount
				/// @returns	A new vector
				//=====================================================================================================
				
				template<class ST> inline Vector4<T> operator * ( ST scale ) const
				{
					return Vector4<T>
					(
						(T)( this->x * scale ), 
						(T)( this->y * scale ),
						(T)( this->z * scale ),
						(T)( this->w * scale )
					);
				}
				
				//=====================================================================================================
				/// @brief		Scalar addition. Returns a new vector with the given value added to it.
				/// @param		value - amount to add
				/// @returns	A new vector with the given amount added
				//=====================================================================================================
				
				template<class ST> inline Vector4<T> operator + ( ST value ) const
				{
					return Vector4<T>
					(
						(T)( this->x + value ), 
						(T)( this->y + value ),
						(T)( this->z + value ),
						(T)( this->w + value )
					);
				}								
				
				//=====================================================================================================
				/// @brief		Scalar subtraction. Returns a new vector with the given value subtracted from it.
				/// @param		value - amount to subtract
				/// @returns	A new vector with the given amount subtracted
				//=====================================================================================================
				
				template<class ST> inline Vector4<T> operator - ( ST value ) const
				{
					return Vector4<T>
					(
						(T)( this->x - value ), 
						(T)( this->y - value ),
						(T)( this->z - value ),
						(T)( this->w - value )
					);
				}				
				
				//=====================================================================================================
				/// @brief		Scalar division. Returns a new vector divided by the given amount.
				/// @param		scale - scaling amount
				/// @returns	A new vector
				/// @pre		Scale must not be zero				
				//=====================================================================================================
				
				template<class ST> inline Vector4<T> operator / ( ST scale ) const
				{
					ENSURE( scale != 0 , "Scale must not be zero" );
				
					return Vector4<T>
					(
						(T)( this->x / scale ),
						(T)( this->y / scale ),
						(T)( this->z / scale ),
						(T)( this->w / scale )
					);
				}
				
				//=====================================================================================================
				/// @brief		Vector addition. Adds the given vector to this vector.
				/// @param		v - vector to add
				//=====================================================================================================
				
				template<class VT> inline void operator += ( const Vector4<VT> & v )
				{
					this->x = (T)( this->x + v.x );
					this->y = (T)( this->y + v.y );
					this->z = (T)( this->z + v.z );
					this->w = (T)( this->w + v.w );
				}
				
				//=====================================================================================================
				/// @brief		Vector subtraction. Subtracts the given vector from this vector.
				/// @param		v - vector to subtract
				//=====================================================================================================
				
				template<class VT> inline void operator -= ( const Vector4<VT> & v )
				{
					this->x = (T)( this->x - v.x );
					this->y = (T)( this->y - v.y );
					this->z = (T)( this->z - v.z );
					this->w = (T)( this->w - v.w );
				}				
				
				//=====================================================================================================
				/// @brief		Scalar addition. Adds the given scalar to this vector.
				/// @param		value - value to add
				//=====================================================================================================
				
				template<class ST> inline void operator += ( ST value )
				{
					this->x = (T)( this->x + value );
					this->y = (T)( this->y + value );
					this->z = (T)( this->z + value );
					this->w = (T)( this->w + value );
				}
				
				//=====================================================================================================
				/// @brief		Scalar subtraction. Subtracts the given scalar to this vector.
				/// @param		value - value to subtract
				//=====================================================================================================
				
				template<class ST> inline void operator -= ( ST value )
				{
					this->x = (T)( this->x - value );
					this->y = (T)( this->y - value );
					this->z = (T)( this->z - value );
					this->w = (T)( this->w - value );
				}												
				
				//=====================================================================================================
				/// @brief		Scalar product. Scales this vector by the given amount.
				/// @param		scale - scaling amount
				//=====================================================================================================
				
				template<class ST> inline void operator *= ( ST scale )
				{
					this->x = (T)( this->x * scale );
					this->y = (T)( this->y * scale );
					this->z = (T)( this->z * scale );
					this->w = (T)( this->w * scale );
				}		
				
				//=====================================================================================================
				/// @brief		Scalar division. Scales this vector by the given amount.
				/// @param		scale - scaling amount
				/// @pre		Scale must not be zero	
				//=====================================================================================================
				
				template<class ST> inline void operator /= ( ST scale )
				{
					ENSURE( scale != 0 , "Scale must not be zero" );
				
					this->x = (T)( this->x / scale );
					this->y = (T)( this->y / scale );
					this->z = (T)( this->z / scale );
					this->w = (T)( this->w / scale );
				}
				
				//=====================================================================================================
				/// @brief		Unary negation operator. Negates this vector.
				/// @returns	A new vector that is negated
				//=====================================================================================================
				
				inline Vector4<T> operator - ( void ) const
				{
					return Vector4<T>(-x,-y,-z,-w);
				}
				
				//=====================================================================================================
				/// @brief		Unary addition operator. Does nothing
				/// @returns	This vector
				//=====================================================================================================
				
				inline Vector4<T> operator + ( void ) const
				{
					return * this;
				}	
				
				//=====================================================================================================
				/// @brief		Dot product. Returns the dot product with the given vector.
				/// @param		v - vector to do a dot product with
				//=====================================================================================================
				
				template<class VT> inline T Dot ( Vector4<VT> v ) const
				{
					return (T)
					(
						( this->x * v.x )
						+ 
						( this->y * v.y )
						+ 
						( this->z * v.z )
						+ 
						( this->w * v.w )						
					);
				}				
				
				//=====================================================================================================
				/// @brief		Gets the squared length of this vector.
				/// @returns	Squared length of the vector
				/// @remarks	Squared length is returned in the same format as the data type of the vector		
				//=====================================================================================================
				
				inline T SqLength() const
				{
					// Return square length:
					
					return ( this->x * this->x ) + ( this->y * this->y ) + ( this->z * this->z ) + ( this->w * this->w );
				}
				
				//=====================================================================================================
				/// @brief		Gets the length of this vector.
				/// @returns	Length of the vector
				/// @remarks	Length is returned in the same format as the data type of the vector		
				//=====================================================================================================
				
				inline T Length() const
				{
					// Get square length:
					
					T sq_length = SqLength();
					
					// Get the square root of it and return: 
					
					return (T) sqrt(sq_length);
				}
				
				//=====================================================================================================
				/// @brief		Normalises the vector and returns the length of the vector in the process.
				/// @returns	Length of the vector
				/// @remarks	A check for division by zero is done
				//=====================================================================================================
				
				inline T Normalize()
				{
					// Get length
					
					T length = Length();
					
					// Normalize if not zero:
					
					if ( length != (T)(0) )
					{
						x /= length;
						y /= length;
						z /= length;
						w /= length;
					}
					
					// Return the length of the vector
					
					return length;
				}
				
				//=====================================================================================================
				/// @brief		Normalises the vector using approximate calculations
				/// @remarks	No check for division by zero is done. Instead a constant value is added to avoid
				///				division by zero. This will never return length 0!
				//=====================================================================================================
				
				inline void FastNormalize()
				{
					// Calculate reciprocal sqrt of square length
					
					float rsqrt = Float::FastInvSqrt( SqLength() + 0.000001f );
					
					// Normalise:
					
					x *= rsqrt;
					y *= rsqrt;
					z *= rsqrt;
					w *= rsqrt;
				}				
				
				//=====================================================================================================
				/// @brief		Reverses this vector, by negating all elements
				/// @returns	A new vector, which is the reverse of this vector.
				//=====================================================================================================				
				
				inline Vector4<T> Reverse() const
				{
					// Return a negated version of this:
					
					return Vector4<T>
					(
						-this->x,
						-this->y,
						-this->z,
						-this->w
					);
				}
				
				//=====================================================================================================
				/// @brief		Returns the normalized direction of this vector.
				/// @returns	Normalized direction of the vector.
				//=====================================================================================================												
				
				inline Vector4<T> Direction() const
				{
					// Make a copy of this vector:
					
					Vector4<T> v(this);
					
					// Normalise:
					
					v.Normalize();
					
					// Return it:
					
					return v;
				}
				
				//=====================================================================================================
				/// @brief		Tells if the two vectors are exactly equal
				/// @param		v - other vector to compare to
				/// @returns	True if the two vectors are exactly equal
				//=====================================================================================================												
				
				template<class VT> inline bool operator == ( const Vector4<VT> & v ) const
				{
					// See if equal:
					
					if 
					( 
						(T) v.x == this->x &&
						(T) v.y == this->y &&
						(T) v.z == this->z &&
						(T) v.w == this->w
					)
					{
						return true;
					}
					else
					{
						return false;
					}
				}	
				
				//=====================================================================================================
				/// @brief		Tells if the two vectors are exactly not equal
				/// @param		v - other vector to compare to
				/// @returns	True if the two vectors are exactly not equal
				/// @remarks	Internally this function uses the precision or data type of this vector.
				//=====================================================================================================												
				
				template<class VT> inline bool operator != ( const Vector4<VT> & v ) const
				{
					// See if not equal:
					
					if 
					( 
						(T) v.x != this->x ||
						(T) v.y != this->y ||
						(T) v.z != this->z ||
						(T) v.w != this->w
					)
					{
						return true;
					}
					else
					{
						return false;
					}
				}
				
				//=====================================================================================================
				/// @brief		Tells if the length of this vector is greater than the length of the given vector
				/// @param		v - other vector to compare to
				/// @returns	True if this vectors length is greater than the other vectors length
				/// @remarks	Internally this function uses the precision or data type of this vector.
				//=====================================================================================================
				
				template<class VT> inline bool operator > ( const Vector4<VT> & v ) const
				{
					// Get the two lengths:
					
					T l1 = SqLength(); T l2 = (T) v.SqLength();
					
					// Return result:
					
					return l1 > l2;
				}		
				
				//=====================================================================================================
				/// @brief		Tells if the length of this vector is greater than or equal to the length of the given vector
				/// @param		v - other vector to compare to
				/// @returns	True if this vectors length is greater than or equal to the other vectors length
				/// @remarks	Internally this function uses the precision or data type of this vector.
				//=====================================================================================================
				
				template<class VT> inline bool operator >= ( const Vector4<VT> & v ) const
				{
					// Get the two lengths:
					
					T l1 = SqLength(); T l2 = (T) v.SqLength();
					
					// Return result:
					
					return l1 >= l2;
				}
				
				//=====================================================================================================
				/// @brief		Tells if the length of this vector is less than the length of the given vector
				/// @param		v - other vector to compare to
				/// @returns	True if this vectors length is less than the other vectors length
				/// @remarks	Internally this function uses the precision or data type of this vector.
				//=====================================================================================================
				
				template<class VT> inline bool operator < ( const Vector4<VT> & v ) const
				{
					// Get the two lengths:
					
					T l1 = SqLength(); T l2 = (T) v.SqLength();
					
					// Return result:
					
					return l1 < l2;
				}
				
				//=====================================================================================================
				/// @brief		Tells if the length of this vector is less than or equal to the length of the given vector
				/// @param		v - other vector to compare to
				/// @returns	True if this vectors length is less than or equal to the other vectors length
				/// @remarks	Internally this function uses the precision or data type of this vector.
				//=====================================================================================================
				
				template<class VT> inline bool operator <= ( const Vector4<VT> & v ) const
				{
					// Get the two lengths:
					
					T l1 = SqLength(); T l2 = (T) v.SqLength();
					
					// Return result:
					
					return l1 <= l2;
				}
				
				//=====================================================================================================
				/// @brief		Compares two vectors for equality with a given margin for error.
				/// @param		v - other vector to compare to
				/// @param		error_margin - margin for error. this is the maximum distance in length between the two 
				///				vectors.
				/// @remarks	This is slower than a normal compare, but can be useful for situations where a margin
				///				of error is required.
				//=====================================================================================================
				
				template<class VT, class ET> inline bool Equals ( const Vector4<VT> & v , ET error_margin = 0.0001f ) const
				{
					// Subtract the two vectors:
					
					Vector4<T> sub =
					(
						this->x - (T) v.x,
						this->y - (T) v.y,
						this->z - (T) v.z,
						this->w - (T) v.w
					);
					
					// Get the square length of it:
					
					ET sq_length = (ET) sub.SqLength();
					
					// See if less than or equal to error margin:
					
					if ( sq_length <= error_margin * error_margin )
					{
						// Equal within margin for error:
						
						return true;
					}
					else
					{
						// Not equal:
						
						return false;
					}
				}
				
				//=====================================================================================================
				/// @brief		Casts the vector to a 3d vector, discarding w
				/// @returns	A Vector3 object
				//=====================================================================================================
				
				inline Vector3<T> ToVec3() const
				{
					return Vector3<T>(x,y,z);
				}				
				
				//=====================================================================================================
				/// @brief		Casts the vector to a 2d vector, discarding z and w
				/// @returns	A Vector2 object
				//=====================================================================================================
				
				inline Vector2<T> ToVec2() const
				{
					return Vector2<T>(x,y);
				}
				
				//=====================================================================================================
				/// @brief		Casts the vector to a directx vector2, discarding z and w
				/// @returns	A direct x vector 2
				//=====================================================================================================				
				
				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
					inline D3DXVECTOR2 ToDxVec2() const
					{
						return D3DXVECTOR2( x , y );
					}
				
				#endif			
				
				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Casts the vector to a directx vector3, discarding w
				/// @returns	A direct x vector 3
				//=====================================================================================================				
				
				inline D3DXVECTOR3 ToDxVec3() const
				{
					return D3DXVECTOR3( x , y , z );
				}
					
				#endif		
				
				#ifdef GAME_SUPPORT_DIRECT_X_TYPES
				
				//=====================================================================================================
				/// @brief		Casts the vector to a directx vector4
				/// @returns	A direct x vector 4
				//=====================================================================================================				
				
				inline D3DXVECTOR4 ToDxVec4() const
				{
					return D3DXVECTOR4( x , y , z , w );
				}
					
				#endif
				
				//=====================================================================================================
				/// @brief		Converts the vector into array format, saving it to the given array
				/// @param		a - array to save to
				/// @pre		Given array cannot be null
				//=====================================================================================================
				
				template <class AT> inline void ToArray( AT * a ) const
				{
					ENSURE(a,"Given array cannot be null");
					
					a[0] = (AT) x;
					a[1] = (AT) y;
					a[2] = (AT) z;
					a[3] = (AT) w;
				}
				
				//=====================================================================================================
				/// @brief		Returns a vector with all components set to zero.
				/// @returns	A vector with all components set to zero.
				//=====================================================================================================
				
				static inline Vector4<T> Zero()
				{
					return Vector4<T>
					(
						(T) 0 ,
						(T) 0 , 
						(T) 0 ,
						(T) 0
					);
				}
				
				//=====================================================================================================
				/// @brief		Returns a vector with all components set to one.
				/// @returns	A vector with all components set to one.
				//=====================================================================================================
				
				static inline Vector4<T> One()
				{
					return Vector4<T>
					(
						(T) 1 ,
						(T) 1 ,
						(T) 1 ,
						(T) 1 
					);
				}
				
				//=====================================================================================================
				/// @brief		Returns a vector with with the x component set to one.
				/// @returns	A vector with the x component set to 1.
				//=====================================================================================================
				
				static inline Vector4<T> UnitX()
				{
					return Vector4<T>
					(
						(T) 1 ,
						(T) 0 ,
						(T) 0 , 
						(T) 0
					);
				}
				
				//=====================================================================================================
				/// @brief		Returns a vector with with the y component set to one.
				/// @returns	A vector with the y component set to 1.
				//=====================================================================================================
				
				static inline Vector4<T> UnitY()
				{
					return Vector4<T>
					(
						(T) 0 ,
						(T) 1 , 
						(T) 0 ,
						(T) 0 
					);
				}
				
				//=====================================================================================================
				/// @brief		Returns a vector with with the z component set to one.
				/// @returns	A vector with the z component set to 1.
				//=====================================================================================================
				
				static inline Vector4<T> UnitZ()
				{
					return Vector4<T>
					(
						(T) 0 ,
						(T) 0 , 
						(T) 1 ,
						(T) 0
					);
				}
				
				//=====================================================================================================
				/// @brief		Returns a vector with with the w component set to one.
				/// @returns	A vector with the w component set to 1.
				//=====================================================================================================
				
				static inline Vector4<T> UnitW()
				{
					return Vector4<T>
					(
						(T) 0 ,
						(T) 0 , 
						(T) 0 ,
						(T) 1
					);
				}
				
				//=====================================================================================================
				/// @brief		Returns the distance to another point
				/// @param		v - other point to get the distance to
				/// @returns	Distance to the given point
				/// @remarks	This returns four dimensional distance. If 3d distance is needed, convert to a 
				///				three dimensional vector first.
				//=====================================================================================================
				
				template <class VT> inline T DistanceTo( const Vector4<VT> & v ) const
				{
					// Subtract the two points
					
					T xs = (T)(v.x) - x;
					T ys = (T)(v.y) - y;
					T zs = (T)(v.z) - z;
					T ws = (T)(v.w) - w;
					
					// Get the square distance:
					
					T sq_dist = xs * xs + ys * ys + zs * zs + ws * ws;
					
					// Now get the square root of it and return:
					
					return Float::Sqrt(sq_dist);
				}
				
				//=====================================================================================================
				/// @brief		Returns the square distance to another point
				/// @param		v - other point to get the distance to
				/// @returns	Distance to the given point
				/// @remarks	This returns four dimensional square distance. If 3d distance is needed, convert to a 
				///				three dimensional vector first.				
				//=====================================================================================================
				
				template <class VT> inline T SqDistanceTo( const Vector4<VT> & v ) const
				{
					// Subtract the two points
					
					T xs = (T)(v.x) - x;
					T ys = (T)(v.y) - y;
					T zs = (T)(v.z) - z;
					T ws = (T)(v.w) - w;
					
					// Get the square distance:
					
					T sq_dist = xs * xs + ys * ys + zs * zs + ws * ws;
					
					// Return the square distance
					
					return sq_dist;
				}				
		
			//=========================================================================================================
			// Variables
			//=========================================================================================================
			
				/// X component for the vector
				
				T x;
				
				/// Y component for the vector
				
				T y;

				/// Z component for the vector
				
				T z;
				
				/// W component for the vector
				
				T w;								
	};
};

#endif 
