/*
	Trig Wars.
	Copyright (c) 2008, Darragh Coy.
	All rights reserved.
	Original work(s), source code & author(s) details available at: http://www.darraghcoy.com

	Redistribution and use in source and binary forms, with or without modification, are permitted 
	provided that the following conditions are met:

		* 	Redistributions of source code must retain the above copyright notice, this list of 
			conditions and the following disclaimer.
		* 	Redistributions in binary form must reproduce the above copyright notice, this list of 
			conditions and the following disclaimer in the documentation.
		* 	Neither the name of the author nor the names of contributors may be used to endorse or 
			promote products derived from this software without specific prior written permission.
		* 	The use of any third party libraries, modules or data files provided with the source 
			form of the program must follow and respect the individual licensing terms of these 
			components. You do not have permission to modify or circumvent the licensing terms for 
			provided third party components unless the third party licenses specifically allow this. 
			If you do not agree to the terms of provided third party components then you do not have 
			permission to use them in your derived work(s).
		* 	The terms of this license do not apply to separate (but related) distributions if supplied 
			by the author(s), only to the contents of this source distribution, derived works, and the 
			binaries produced by either modified or unmodified versions of the source code. The terms 
			of this license do not also apply to any third party libraries or components provided with 
			the source distribution. Such third party elements are licensed under their own separate 
			agreements.

	THIS SOFTWARE IS PROVIDED BY DARRAGH COY ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
	BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL DARRAGH COY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
	IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

//=====================================================================================================================
/// @brief		This class is used as a fast temporary storage area for data. The main reason this has been 
///				created (rather than use the regular stack) is because it allows for a variable amount of parameters
///				to be pushed onto it- making it perfect for temporary lists of objects without the overhead or 
///				loss of cache coherency that the 'new' operator brings. It also allows for temporary variables 
///				to be returned from functions, unlike the regular stack.
/// @remarks	In debug mode the stack will validate usage of the stack, however in release this will not happen.
///				All entries shall be aligned to 4 byte boundaries for maximum read/write efficiency.
/// @post		The given memory will be allocated or the program will bail
//=====================================================================================================================

#include "Precompiled.h"

using namespace Game;

//=====================================================================================================================
/// @brief		Creates the stack with the specified size
/// @pre		The given stack size cannot be zero
//=====================================================================================================================

Stack::Stack( unsigned size )
{
	// Make sure ok:
	
	ENSURE( size != 0 , "The given stack size cannot be zero" );
	
	// Figure out how much we are going to alloc: alloc 4 extra bytes so we can do alignment
	
	unsigned alloc_bytes = size + 4;
	
	// Do the allocation:
	
	m_stack_memory = new char[ alloc_bytes];
	
	// If it failed to make the stack then abort:
	
	if ( m_stack_memory == 0 )
	{
		// Do a break in debug mode:
		
		#ifdef DEBUG
		
			DebugBreak();
		
		#endif
		
		// Now issue an error
		
		Errors::IssueError( "Out of memory. Unable to create requested stack." );
	}
	
	// Calculate the aligned address of the memory:
	
	m_stack_memory_aligned = (void*)((((unsigned)(m_stack_memory)) + 3 ) & 0xfffffffc );
	
	// Set the current pointer:
	
	m_current = m_stack_memory_aligned;
	
	// Set the end pointer:
	
	m_end = (void*)((unsigned)(m_stack_memory_aligned) + size);
}

//=====================================================================================================================
/// @brief		Destroys the stack. 
/// @pre		The stack's current pointer should match the start pointer. I.E there should be no allocations left
///				on the stack.
//=====================================================================================================================

Stack::~Stack()
{
	// Make sure the stack pointer is ok:
	
	ENSURE( m_stack_memory_aligned ==  m_current , "The stack's current pointer should match the start pointer. I.E there should be no allocations left on the stack." );
	
	// Dealloc the stack memory:
	
	delete m_stack_memory;
}
