/*
	Trig Wars.
	Copyright (c) 2008, Darragh Coy.
	All rights reserved.
	Original work(s), source code & author(s) details available at: http://www.darraghcoy.com

	Redistribution and use in source and binary forms, with or without modification, are permitted 
	provided that the following conditions are met:

		* 	Redistributions of source code must retain the above copyright notice, this list of 
			conditions and the following disclaimer.
		* 	Redistributions in binary form must reproduce the above copyright notice, this list of 
			conditions and the following disclaimer in the documentation.
		* 	Neither the name of the author nor the names of contributors may be used to endorse or 
			promote products derived from this software without specific prior written permission.
		* 	The use of any third party libraries, modules or data files provided with the source 
			form of the program must follow and respect the individual licensing terms of these 
			components. You do not have permission to modify or circumvent the licensing terms for 
			provided third party components unless the third party licenses specifically allow this. 
			If you do not agree to the terms of provided third party components then you do not have 
			permission to use them in your derived work(s).
		* 	The terms of this license do not apply to separate (but related) distributions if supplied 
			by the author(s), only to the contents of this source distribution, derived works, and the 
			binaries produced by either modified or unmodified versions of the source code. The terms 
			of this license do not also apply to any third party libraries or components provided with 
			the source distribution. Such third party elements are licensed under their own separate 
			agreements.

	THIS SOFTWARE IS PROVIDED BY DARRAGH COY ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
	BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL DARRAGH COY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
	IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

//=====================================================================================================================
/// @brief		This is a simple extension of the ordinary bullet manager that caches references to AI object
///				managers for the purposes of collision detection.
//=====================================================================================================================

#include "Precompiled.h"
#include "Game\PlayerBulletManager.h"
#include "GameObjects\EnemyWaspManager.h"
#include "GameObjects\EnemyFactoryManager.h"
#include "GameObjects\EnemyPredatorManager.h"
#include "Game\GameCore.h"
#include "Resources\TextureManager.h"

using namespace Game;

//=====================================================================================================================
/// @brief		Creates the bullet manager
/// @param		player_ship - the player ship that owns this manager
/// @param		bullet_count - number of bullets to make
/// @param		stack - Stack that holds a pointer for each bullet. This function expects the list of 
///				bullet pointers to be at the very top of the stack.
/// @param		bullet_color - color of each bullet when rendering
/// @param		bullet_length - how long the bullets should be when drawing
/// @pre		Player ship must not be null
/// @pre		Bullet count must be 1 or more
/// @pre		The game must exist
/// @pre		The game must have a texture manager
/// @remarks	The manager takes ownership of bullet memory after this function is called, and is 
///				reponsible for deleting the created bullets.
//=====================================================================================================================

PlayerBulletManager::PlayerBulletManager( PlayerShip * player_ship , unsigned bullet_count , Stack & stack , Color bullet_color , float bullet_length )
: 
BulletManager
(
	bullet_count,
	stack,
	bullet_color,
	bullet_length
)
{
	// Make sure ok:
	
	ENSURE( player_ship != 0								, "Player ship must not be null"			)
	ENSURE( bullet_count >= 1								, "Bullet count must be 1 or more"			)
	ENSURE( GameCore::GetSingleton()						, "The game must exist"						)
	ENSURE( GameCore::GetSingleton()->GetTextureManager()	, "The game must have a texture manager"	)
	
	// Save:
	
	m_player_ship = player_ship;

	// We haven't found these objects yet

	m_factory_mgr = 0; m_wasp_mgr = 0; m_predator_mgr = 0;
	
	// Load our particle texture
	
	m_particle_texture = GameCore::GetSingleton()->GetTextureManager()->GetTexture("Textures\\GlowParticle.dds");
}

//=====================================================================================================================
/// @brief		Destructor. Frees resources
//=====================================================================================================================

PlayerBulletManager::~PlayerBulletManager()
{
	// Free reference to the wasp manager:
	
	if ( m_wasp_mgr ) m_wasp_mgr->RemoveReference();
	
	// Free reference to the factory manager:
	
	if ( m_factory_mgr ) m_factory_mgr->RemoveReference();
	
	// Free reference to the predator manager:
	
	if ( m_predator_mgr ) m_predator_mgr->RemoveReference();	
}

//=====================================================================================================================
/// @brief		Update all function: updates references to AI object managers as well as doing base class logic
/// @pre		The game must exist
/// @pre		The game must have a game object manager
/// @post		All cached AI object managers will be of the correct type
//=====================================================================================================================

void PlayerBulletManager::UpdateAll( float time_delta )
{
	// Make sure ok:
	
	ENSURE( GameCore::GetSingleton()						, "The game must exist"							);
	ENSURE( GameCore::GetSingleton()->GetObjectManager()	, "The game must have a game object manager"	);

	// See if we have a wasp manager:
	
	if ( m_wasp_mgr )
	{
		// See if now out of the game: if so then free
		
		if ( m_wasp_mgr->HasManager() == false )
		{
			m_wasp_mgr->RemoveReference(); m_wasp_mgr = 0;
		}
	}
	else
	{
		// Try to find the enemy wasp manager
		
		m_wasp_mgr = GameCore::GetSingleton()->GetObjectManager()->FindByType<EnemyWaspManager>("EnemyWaspManager");
		
		// Add a reference to it if found
		
		if ( m_wasp_mgr ) m_wasp_mgr->AddReference();
	}
	
	// See if we have a factory manager:
	
	if ( m_factory_mgr )
	{
		// See if now out of the game: if so then free
		
		if ( m_factory_mgr->HasManager() == false )
		{
			m_factory_mgr->RemoveReference(); m_factory_mgr = 0;
		}
	}
	else
	{
		// Try to find the enemy wasp manager
		
		m_factory_mgr = GameCore::GetSingleton()->GetObjectManager()->FindByType<EnemyFactoryManager>("EnemyFactoryManager");
		
		// Add a reference to it if found
		
		if ( m_factory_mgr )  m_factory_mgr->AddReference();
	}	
	
	// See if we have a predator manager:
	
	if ( m_predator_mgr )
	{
		// See if now out of the game: if so then free
		
		if ( m_predator_mgr->HasManager() == false )
		{
			m_predator_mgr->RemoveReference(); m_factory_mgr = 0;
		}
	}
	else
	{
		// Try to find the enemy wasp manager
		
		m_predator_mgr = GameCore::GetSingleton()->GetObjectManager()->FindByType<EnemyPredatorManager>("EnemyPredatorManager");
		
		// Add a reference to it if found
		
		if ( m_predator_mgr ) m_predator_mgr->AddReference();
	}	

	// Do base class update:
	
	BulletManager::UpdateAll(time_delta);
}

//=====================================================================================================================
/// @brief		Annouces the type name for this object.
/// @returns	Type name of this object
//=====================================================================================================================

const char * PlayerBulletManager::GetTypeName() const
{
	return "PlayerBulletManager";
}
