/*
	Trig Wars.
	Copyright (c) 2008, Darragh Coy.
	All rights reserved.
	Original work(s), source code & author(s) details available at: http://www.darraghcoy.com

	Redistribution and use in source and binary forms, with or without modification, are permitted 
	provided that the following conditions are met:

		* 	Redistributions of source code must retain the above copyright notice, this list of 
			conditions and the following disclaimer.
		* 	Redistributions in binary form must reproduce the above copyright notice, this list of 
			conditions and the following disclaimer in the documentation.
		* 	Neither the name of the author nor the names of contributors may be used to endorse or 
			promote products derived from this software without specific prior written permission.
		* 	The use of any third party libraries, modules or data files provided with the source 
			form of the program must follow and respect the individual licensing terms of these 
			components. You do not have permission to modify or circumvent the licensing terms for 
			provided third party components unless the third party licenses specifically allow this. 
			If you do not agree to the terms of provided third party components then you do not have 
			permission to use them in your derived work(s).
		* 	The terms of this license do not apply to separate (but related) distributions if supplied 
			by the author(s), only to the contents of this source distribution, derived works, and the 
			binaries produced by either modified or unmodified versions of the source code. The terms 
			of this license do not also apply to any third party libraries or components provided with 
			the source distribution. Such third party elements are licensed under their own separate 
			agreements.

	THIS SOFTWARE IS PROVIDED BY DARRAGH COY ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
	BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL DARRAGH COY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
	IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef DAMAGEABLE_OBJECT_H
#define DAMAGEABLE_OBJECT_H

//=====================================================================================================================
//=====================================================================================================================

namespace Game
{
	//=================================================================================================================
	/// @brief		Base class for any object that has health and that can be killed.
	/// @remarks	This is not a game object type in itself, rather an interface that can be inherited by any object.
	//=================================================================================================================

	class DamageableObject
	{
		public:

			//=========================================================================================================
			// Functions
			//=========================================================================================================
			
				//=====================================================================================================
				/// @brief		Constructor: Sets health to the maximum
				/// @pre		Given health must be greater than zero
				//=====================================================================================================
				
				inline DamageableObject( float max_health )
				{
					// Make sure ok:
				
					ENSURE( max_health > 0 , "Given health must be greater than zero" );
					
					// Set health to the max:
					
					m_health		= max_health;
					m_max_health	= max_health;
				}
				
				//=====================================================================================================
				/// @brief		Heals the object by the given amount
				/// @param		amount - amount to heal by
				//=====================================================================================================
				
				inline void Heal( float amount )
				{
					// Add to health:
					
					m_health += amount;
					
					// Make sure not past the maximum
					
					if ( m_health > m_max_health ) m_health = m_max_health;
				}
				
				//=====================================================================================================
				/// @brief		Damages the object by the given amount
				/// @param		amount - amount to damage by
				//=====================================================================================================
				
				inline void Damage( float amount )
				{
					// See if dead already:
					
					if ( IsDead() == false )
					{				
						// Remove from health:
						
						m_health -= amount;												
						
						// Make sure not past 0
						
						if ( m_health < 0 ) m_health = 0;
					}
				}
				
				//=====================================================================================================
				/// @brief		Tells if the object is 'Alive' or has health left
				/// @returns	True if the object is alive
				//=====================================================================================================
				
				inline bool IsAlive(){ return m_health > 0; }
				
				//=====================================================================================================
				/// @brief		Tells if the object is 'Dead' or has no health left
				/// @returns	True if the object is dead
				//=====================================================================================================
				
				inline bool IsDead(){ return m_health <= 0; }
				
				//=====================================================================================================
				/// @brief		Restores the object's health to the maximum
				//=====================================================================================================
								
				inline void Revive(){ m_health = m_max_health; }						

				//=====================================================================================================
				/// @brief		Sets the object's health to zero, thereby killing the object
				//=====================================================================================================
								
				inline void Kill()
				{ 
					if ( IsDead() == false )
					{ 				
						m_health = 0;
					}
				}

				//=====================================================================================================
				/// @brief		Gets the current amount of health the object has
				/// @returns	Current health the object has
				//=====================================================================================================
								
				inline float GetHealth(){ return m_health; }
				
				//=====================================================================================================
				/// @brief		Gets the maximum amount of health the object can have
				/// @returns	Maximum amount of health the object can have
				//=====================================================================================================
								
				inline float GetMaxHealth(){ return m_max_health; }
				
		private:

			//=========================================================================================================
			// Variables
			//=========================================================================================================

				/// Health the object has
				
				float m_health;
				
				/// Maximum amount of health the object can have
				
				float m_max_health;
	};
};

#endif
