/*
	Trig Wars.
	Copyright (c) 2008, Darragh Coy.
	All rights reserved.
	Original work(s), source code & author(s) details available at: http://www.darraghcoy.com

	Redistribution and use in source and binary forms, with or without modification, are permitted 
	provided that the following conditions are met:

		* 	Redistributions of source code must retain the above copyright notice, this list of 
			conditions and the following disclaimer.
		* 	Redistributions in binary form must reproduce the above copyright notice, this list of 
			conditions and the following disclaimer in the documentation.
		* 	Neither the name of the author nor the names of contributors may be used to endorse or 
			promote products derived from this software without specific prior written permission.
		* 	The use of any third party libraries, modules or data files provided with the source 
			form of the program must follow and respect the individual licensing terms of these 
			components. You do not have permission to modify or circumvent the licensing terms for 
			provided third party components unless the third party licenses specifically allow this. 
			If you do not agree to the terms of provided third party components then you do not have 
			permission to use them in your derived work(s).
		* 	The terms of this license do not apply to separate (but related) distributions if supplied 
			by the author(s), only to the contents of this source distribution, derived works, and the 
			binaries produced by either modified or unmodified versions of the source code. The terms 
			of this license do not also apply to any third party libraries or components provided with 
			the source distribution. Such third party elements are licensed under their own separate 
			agreements.

	THIS SOFTWARE IS PROVIDED BY DARRAGH COY ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
	BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL DARRAGH COY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
	IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

//=====================================================================================================================
/// @brief		Represents a offscreen 2D render target surface.
/// @remarks	Render target data is volatile and will be lost after a device reset. 
//=====================================================================================================================

#include "Precompiled.h"
#include "Graphics\RenderTarget.h"

using namespace Game;

//=====================================================================================================================
/// @brief		Creates the render target with the given size and format
/// @param		display - display to use for the render target
/// @param		size - size of the render target
/// @param		format - format of the render target
/// @param		aa_level - level of anti-aliasing to use for the render target
/// @param		clamp_to_max_texure_size - if true then the render target will automatically clamp the size of the 
///				render target to the maximum texture size. Otherwise if false, then the creation may fail if the 
///				requested size is too big for the hardware.
/// @pre		The given size must be valid
/// @remarks	The given multi-sample quality may or may not be used depending on hardware support.
///				The render target will attempt to best match the given setting however.
//=====================================================================================================================

RenderTarget::RenderTarget( Display * display , const Vec2u & size , Format format , Display::AntialiasingLevel aa_level , bool clamp_to_max_texure_size )
:
DisplayEvictable(display)
{
	// Make sure ok:
	
	ENSURE( size.x > 0 , "The given size must be valid" )
	ENSURE( size.y > 0 , "The given size must be valid" )

	// Initialiase everything:
	
	m_surface_format	= format;
	m_size				= size;
	m_aa_level			= aa_level;
	m_surface			= 0;
	m_texture			= 0;
	
	// See if we are to validate the size:
	
	if ( clamp_to_max_texure_size )
	{
		if ( m_size.x > GetDisplay()->GetMaxTextureWidth() )
		{
			m_size.x = GetDisplay()->GetMaxTextureWidth();
		}
		
		if ( m_size.y > GetDisplay()->GetMaxTextureHeight() )
		{
			m_size.y = GetDisplay()->GetMaxTextureHeight();
		}		
	}	
	
	// Recreate the render target:
	
	RecreateEvictableResources();
}

//=====================================================================================================================
/// @brief		Destructor: releases the render target
//=====================================================================================================================

RenderTarget::~RenderTarget()
{
	// Release our resources if valid:
	
	if ( IsValid() )
	{
		RenderTarget::ReleaseEvictableResources();
	}
}

//=====================================================================================================================
/// @brief		Annouces the type name for this object.
/// @returns	Type name of this object
//=====================================================================================================================

const char * RenderTarget::GetTypeName() const
{
	return "RenderTarget";
}

//=====================================================================================================================
/// @brief		Releases the render target before a device reset: not to be called by anything other than the display!
/// @pre		The render target must be valid
//=====================================================================================================================

void RenderTarget::ReleaseEvictableResources()
{
	// Make sure ok:

	ENSURE( m_is_valid , "The render target must be valid" )
	
	// Release our render target surface:
	
	m_surface->Release(); m_surface = 0;
	
	// Release our texture:
	
	m_texture->Release(); m_texture = 0;
	
	// No longer valid:
	
	m_is_valid = false;
}

//=====================================================================================================================
/// @brief		Recreates the render target after a device reset: not to be called by anything other than the display!
/// @pre		The render target must have no resources on hand before the call
//=====================================================================================================================

void RenderTarget::RecreateEvictableResources()
{
	// Make sure ok:

	ENSURE( m_surface == 0 && m_texture == 0 && m_is_valid == false , "The render target must have no resources on hand before the call" )
	
	// Determine the maximum aa level:
	
	Display::AntialiasingLevel max_aa_level = GetMaxAALevel();
	
	// See if we are above it:
	
	if ( (unsigned)(m_aa_level) > (unsigned)(m_aa_level) )
	{
		// Use the max aa level instead:
		
		m_aa_level = max_aa_level;
	}
	
	// Get the direct3d device:
	
	IDirect3DDevice9 * device = GetDisplay()->GetDevice();
	
	// Get the number of quality levels for this aa level.. What the hell is this about?
	
	DWORD quality_levels = 0; 
	
	GetDisplay()->GetDirect3D()->CheckDeviceMultiSampleType
	(
		D3DADAPTER_DEFAULT						,
		D3DDEVTYPE_HAL							,
		(D3DFORMAT) m_surface_format			,
		GetDisplay()->IsFullscreen() == false	,
		(D3DMULTISAMPLE_TYPE) m_aa_level		,
		&quality_levels
	);
	
	// Now make the render target surface:
	
	HRESULT result = device->CreateRenderTarget
	(
		m_size.x							,
		m_size.y							,
		(D3DFORMAT)	m_surface_format		,
		(D3DMULTISAMPLE_TYPE) m_aa_level	,
		quality_levels - 1					,
		false								,	// not lockable
		&m_surface							,
		0										// shared handle: only on vista
	);
	
	// See if that was ok:
	
	if ( result == D3D_OK && m_surface )
	{
		// We are ok so far: try to create a texture for our render target that we can copy the results into
		
		result = device->CreateTexture
		(
			m_size.x						,
			m_size.y						,
			1								,	
			D3DUSAGE_RENDERTARGET			,
			(D3DFORMAT) m_surface_format	,
			D3DPOOL_DEFAULT					,
			&m_texture						,
			0
		);
		
		// See if that was ok:
		
		if ( result == D3D_OK && m_texture )
		{
			// All good:
			
			m_is_valid = true;
		}
		else
		{
			// Fail: release our render target
			
			m_surface->Release(); m_surface = 0;
			
			// Release the render texture if there:
			
			if ( m_texture ) m_texture->Release();
			
			// Done with this:
			
			m_texture = 0;			
		}
	}	
	else
	{
		// Not ok: if the surface is there then remove it
		
		if ( m_surface ) m_surface->Release();
		
		// Done with this:
		
		m_surface = 0;
	}
}

//=====================================================================================================================
/// @brief		Determines the maximum anti-aliasing level supported by the settings for the render target
/// @returns	The maximum aa level that can be used by this render target
//=====================================================================================================================

Display::AntialiasingLevel RenderTarget::GetMaxAALevel()
{
	// Get the direct3d object:
	
	IDirect3D9 * d3d = GetDisplay()->GetDirect3D();

	// Store the result of level checking here:
	
	HRESULT result = D3D_OK;
	
	// Macro to check the levels:
			
	#define CHECK_AA_LEVEL(AA_LEVEL)			\
												\
	result = d3d->CheckDeviceMultiSampleType	\
	(											\
		D3DADAPTER_DEFAULT	,					\
		D3DDEVTYPE_HAL		,					\
		D3DFMT_X8R8G8B8		,					\
		GetDisplay()->IsFullscreen() == false ,	\
		AA_LEVEL ,								\
		0										\
	);

	// Check 2x:
	
	CHECK_AA_LEVEL(D3DMULTISAMPLE_2_SAMPLES); if ( result != D3D_OK ) return Display::ANTIALIAS_NONE;
	
	// Check 4x:
	
	CHECK_AA_LEVEL(D3DMULTISAMPLE_4_SAMPLES); if ( result != D3D_OK ) return Display::ANTIALIAS_2_SAMPLES;
	
	// Check 8x:
	
	CHECK_AA_LEVEL(D3DMULTISAMPLE_8_SAMPLES); if ( result != D3D_OK ) return Display::ANTIALIAS_4_SAMPLES;	

	// Check 16x:
	
	CHECK_AA_LEVEL(D3DMULTISAMPLE_16_SAMPLES); if ( result != D3D_OK ) return Display::ANTIALIAS_8_SAMPLES;	
	
	// Return 16x:
			
	return Display::ANTIALIAS_16_SAMPLES;	
}

//=====================================================================================================================
/// @brief		Copies the contents of the render target into a texture that can be used: must be called before 
///				using ApplyTexture().
/// @pre		Render target must be valid
/// @remarks	If this is not called before calling ApplyTexture(), the resultant texture will either be undefined
///				or the same texture from the last time UpdateTexture() was called.
//=====================================================================================================================

void RenderTarget::UpdateTexture()
{
	// Make sure ok:

	ENSURE( m_is_valid , "The render target must be valid" )
	
	// Get the top mip map in the render target texture
	
	IDirect3DSurface9 * dest = 0; m_texture->GetSurfaceLevel(0,&dest);
	
	// This should never happen:
	
	ENSURE( dest , "IDirect3DTexture9::GetSurfaceLevel(0,&dest) returned null!" )
	
	// Get the d3d device:
	
	IDirect3DDevice9 * d = GetDisplay()->GetDevice();
	
	// Do the copy:
	
	d->StretchRect( m_surface , 0 , dest , 0 , D3DTEXF_NONE );

	// Done with this:

	dest->Release();
}

//=====================================================================================================================
/// @brief		Applys the render target texture to the given sampler unit. Be sure to
/// @param		sampler_unit - sampler unit to apply the render target to
/// @pre		Render target must be valid
//=====================================================================================================================

void RenderTarget::ApplyTexture( unsigned sampler_unit )
{
	// Make sure ok:

	ENSURE( m_is_valid , "The render target must be valid" )
	
	// Get the d3d device and bind the texture:
	
	GetDisplay()->GetDevice()->SetTexture(sampler_unit,m_texture);
}
