/*
	Trig Wars.
	Copyright (c) 2008, Darragh Coy.
	All rights reserved.
	Original work(s), source code & author(s) details available at: http://www.darraghcoy.com

	Redistribution and use in source and binary forms, with or without modification, are permitted 
	provided that the following conditions are met:

		* 	Redistributions of source code must retain the above copyright notice, this list of 
			conditions and the following disclaimer.
		* 	Redistributions in binary form must reproduce the above copyright notice, this list of 
			conditions and the following disclaimer in the documentation.
		* 	Neither the name of the author nor the names of contributors may be used to endorse or 
			promote products derived from this software without specific prior written permission.
		* 	The use of any third party libraries, modules or data files provided with the source 
			form of the program must follow and respect the individual licensing terms of these 
			components. You do not have permission to modify or circumvent the licensing terms for 
			provided third party components unless the third party licenses specifically allow this. 
			If you do not agree to the terms of provided third party components then you do not have 
			permission to use them in your derived work(s).
		* 	The terms of this license do not apply to separate (but related) distributions if supplied 
			by the author(s), only to the contents of this source distribution, derived works, and the 
			binaries produced by either modified or unmodified versions of the source code. The terms 
			of this license do not also apply to any third party libraries or components provided with 
			the source distribution. Such third party elements are licensed under their own separate 
			agreements.

	THIS SOFTWARE IS PROVIDED BY DARRAGH COY ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
	BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL DARRAGH COY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
	IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef TEXTURE_H
#define TEXTURE_H

//=====================================================================================================================
//=====================================================================================================================

#include "Resources\TextureManager.h"

//=====================================================================================================================
//=====================================================================================================================

namespace Game
{
	//=================================================================================================================
	/// @brief		Represents a loaded texture resource. Internally this class is just storing a managed resource, 
	///				but this is the end user interface for textures so the class comes with additional functionality.
	//=================================================================================================================
	
	class Texture
	{
		public:

			//=========================================================================================================
			// Functions
			//=========================================================================================================
			
				//=====================================================================================================
				/// @brief	Creates a default invalid texture
				//=====================================================================================================
				
				Texture(){}			
	        
				//=====================================================================================================
				///	@brief	Constructor. Creates the texture.
				///	@param	resource - managed resource used by the texture
				//=====================================================================================================
				
				Texture( TextureManager::ManagedResource & resource ) : m_resource(resource){}
				
				//=====================================================================================================
				///	@brief		Returns the managed resource used by the texutre
				///	@returns	Managed resource used by the manager
				//=====================================================================================================
				
				TextureManager::ManagedResource GetResource() const { return m_resource; }
				
				///	Applys the texture to the given sampler unit
				
				void Apply( unsigned sampler_unit ) const;
				
				///	Gets the width of the texture
				
				int GetWidth() const;
				
				///	Gets the height of the texture
				
				int GetHeight() const;							
				
				//=====================================================================================================
				///	@brief		Tells if the resource is valid
				///	@returns	True if the resource is valid
				//=====================================================================================================
				
				inline bool IsValid() const { return m_resource.IsValid(); }			
				
				//=====================================================================================================
				///	@brief		Assignment operator
				/// @param		other - the other texture resource
				/// @returns	this resource
				//=====================================================================================================			
				
				inline Texture & operator = ( const Texture & other )
				{
					// Only if not the same:
					
					if ( &other != this )
					{
						// Do the assignment 
						
						m_resource = other.m_resource;
					}
					
					// Return this resource
					
					return * this;		
				}
				
				//=====================================================================================================
				///	@brief		Copy constructor
				/// @param		other - the other texture resource
				//=====================================================================================================			
				
				Texture( const Texture & other ) : m_resource( other.m_resource ){}				
				
				//=====================================================================================================
				/// @brief		Comparison operator. Compares two textures based on the memory address of the 
				///				texture resource pointed to.
				/// @returns	Result of numeric comparison based on memory address of the resource pointed to.
				//=====================================================================================================
				
				inline bool operator == ( const Texture & other ) const 
				{
					return m_resource.GetResource() == other.m_resource.GetResource();
				}
				
				//=====================================================================================================
				/// @brief		Comparison operator. Compares two textures based on the memory address of the 
				///				texture resource pointed to.
				/// @returns	Result of numeric comparison based on memory address of the resource pointed to.
				//=====================================================================================================
				
				inline bool operator != ( const Texture & other ) const 
				{
					return m_resource.GetResource() != other.m_resource.GetResource();
				}	
				
				//=====================================================================================================
				/// @brief		Comparison operator. Compares two textures based on the memory address of the 
				///				texture resource pointed to.
				/// @returns	Result of numeric comparison based on memory address of the resource pointed to.
				//=====================================================================================================
				
				inline bool operator > ( const Texture & other ) const 
				{
					return m_resource.GetResource() > other.m_resource.GetResource();
				}
				
				//=====================================================================================================
				/// @brief		Comparison operator. Compares two textures based on the memory address of the 
				///				texture resource pointed to.
				/// @returns	Result of numeric comparison based on memory address of the resource pointed to.
				//=====================================================================================================
				
				inline bool operator < ( const Texture & other ) const 
				{
					return m_resource.GetResource() < other.m_resource.GetResource();
				}
				
				//=====================================================================================================
				/// @brief		Comparison operator. Compares two textures based on the memory address of the 
				///				texture resource pointed to.
				/// @returns	Result of numeric comparison based on memory address of the resource pointed to.
				//=====================================================================================================
				
				inline bool operator >= ( const Texture & other ) const 
				{
					return m_resource.GetResource() >= other.m_resource.GetResource();
				}
				
				//=====================================================================================================
				/// @brief		Comparison operator. Compares two textures based on the memory address of the 
				///				texture resource pointed to.
				/// @returns	Result of numeric comparison based on memory address of the resource pointed to.
				//=====================================================================================================
				
				inline bool operator <= ( const Texture & other ) const 
				{
					return m_resource.GetResource() <= other.m_resource.GetResource();
				}
				
		private:

			//=========================================================================================================
			// Variables
			//=========================================================================================================
	        
				/// Texture resource used by the texture
				
				TextureManager::ManagedResource m_resource;
	};	

};

#endif
