/*
	Trig Wars.
	Copyright (c) 2008, Darragh Coy.
	All rights reserved.
	Original work(s), source code & author(s) details available at: http://www.darraghcoy.com

	Redistribution and use in source and binary forms, with or without modification, are permitted 
	provided that the following conditions are met:

		* 	Redistributions of source code must retain the above copyright notice, this list of 
			conditions and the following disclaimer.
		* 	Redistributions in binary form must reproduce the above copyright notice, this list of 
			conditions and the following disclaimer in the documentation.
		* 	Neither the name of the author nor the names of contributors may be used to endorse or 
			promote products derived from this software without specific prior written permission.
		* 	The use of any third party libraries, modules or data files provided with the source 
			form of the program must follow and respect the individual licensing terms of these 
			components. You do not have permission to modify or circumvent the licensing terms for 
			provided third party components unless the third party licenses specifically allow this. 
			If you do not agree to the terms of provided third party components then you do not have 
			permission to use them in your derived work(s).
		* 	The terms of this license do not apply to separate (but related) distributions if supplied 
			by the author(s), only to the contents of this source distribution, derived works, and the 
			binaries produced by either modified or unmodified versions of the source code. The terms 
			of this license do not also apply to any third party libraries or components provided with 
			the source distribution. Such third party elements are licensed under their own separate 
			agreements.

	THIS SOFTWARE IS PROVIDED BY DARRAGH COY ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
	BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL DARRAGH COY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
	IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef KEYBOARD_H
#define KEYBOARD_H

//=====================================================================================================================
//=====================================================================================================================

namespace Game
{
	//=================================================================================================================
	/// @brief		Class representing the keyboard. Provides functions for querying for keyboard input as well as 
	///				Virtual Key to ascii key conversions.
	//=================================================================================================================
	
	class Keyboard
	{
		public:
		
			//=========================================================================================================
			// Enums 
			//=========================================================================================================	
			
				/// Represents a keycode for a key on the keyboard
		
				enum Key
				{
					KB_BACKSPACE	= 0x08,
					KB_TAB			= 0x09,
					KB_ENTER		= 0x0d,
					KB_SHIFT		= 0x10,
					KB_CTRL			= 0x11,
					KB_ALT			= 0x12,
					KB_PAUSE		= 0x13,
					KB_CAPSLOCK		= 0x14,
					KB_ESCAPE		= 0x1b,
					KB_SPACE		= 0x20,
					KB_PAGE_UP		= 0x21,
					KB_PAGE_DOWN	= 0x22,
					KB_END			= 0x23,
					KB_HOME			= 0x24,
					KB_LEFT			= 0x25,
					KB_UP			= 0x26,
					KB_RIGHT		= 0x27,
					KB_DOWN			= 0x28,
					KB_PRINT		= 0x2c,
					KB_INSERT		= 0x2d,
					KB_DELETE		= 0x2e,
					KB_NUM0			= 0x30,
					KB_NUM1			= 0x31,
					KB_NUM2			= 0x32,
					KB_NUM3			= 0x33,
					KB_NUM4			= 0x34,
					KB_NUM5			= 0x35,
					KB_NUM6			= 0x36,
					KB_NUM7			= 0x37,
					KB_NUM8			= 0x38,
					KB_NUM9			= 0x39,
					KB_A			= 0x41,
					KB_B			= 0x42,
					KB_C			= 0x43,
					KB_D			= 0x44,
					KB_E			= 0x45,
					KB_F			= 0x46,
					KB_G			= 0x47,
					KB_H			= 0x48,
					KB_I			= 0x49,
					KB_J			= 0x4a,
					KB_K			= 0x4b,
					KB_L			= 0x4c,
					KB_M			= 0x4d,
					KB_N			= 0x4e,
					KB_O			= 0x4f,
					KB_P			= 0x50,
					KB_Q			= 0x51,
					KB_R			= 0x52,
					KB_S			= 0x53,
					KB_T			= 0x54,
					KB_U			= 0x55,
					KB_V			= 0x56,
					KB_W			= 0x57,
					KB_X			= 0x58,
					KB_Y			= 0x59,
					KB_Z			= 0x5A,
					KB_NUMPAD_0		= 0x60,
					KB_NUMPAD_1		= 0x61,
					KB_NUMPAD_2		= 0x62,
					KB_NUMPAD_3		= 0x63,
					KB_NUMPAD_4		= 0x64,
					KB_NUMPAD_5		= 0x65,
					KB_NUMPAD_6		= 0x66,
					KB_NUMPAD_7		= 0x67,
					KB_NUMPAD_8		= 0x68,
					KB_NUMPAD_9		= 0x69,
					KB_MULTIPLY		= 0x6a,
					KB_ADD			= 0x6b,
					KB_SUBTRACT		= 0x6d,
					KB_DECIMAL		= 0x6e,
					KB_DIVIDE		= 0x6f,
					KB_F1			= 0x70,
					KB_F2			= 0x71,
					KB_F3			= 0x72,
					KB_F4			= 0x73,
					KB_F5			= 0x74,
					KB_F6			= 0x75,
					KB_F7			= 0x76,
					KB_F8			= 0x77,
					KB_F9			= 0x78,
					KB_F10			= 0x79,
					KB_F11			= 0x7a,
					KB_F12			= 0x7b,
					KB_NUM_LOCK		= 0x90,
					KB_SCROLL_LOCK	= 0x91,
					KB_BACK_TICK	= 0xdf				
				};

			//=========================================================================================================
			// Functions
			//=========================================================================================================
		
				/// Query if a keyboard key is pressed
		
				static bool IsKeyPressed( Key key_code );
				
				/// Query if a keyboard key is released
				
				static inline bool IsKeyReleased( Key key_code ){ return ! IsKeyPressed(key_code); }
				
				/// Query if a toggle key such as the capslock or numlock is on
				
				static bool IsKeyToggled( Key key_code );
				
				/// Translates the given virtual key code to its ascii representation.
				
				static char KeyToAscii( Key key_code );
	};
};

#endif
