/*
	Trig Wars.
	Copyright (c) 2008, Darragh Coy.
	All rights reserved.
	Original work(s), source code & author(s) details available at: http://www.darraghcoy.com

	Redistribution and use in source and binary forms, with or without modification, are permitted 
	provided that the following conditions are met:

		* 	Redistributions of source code must retain the above copyright notice, this list of 
			conditions and the following disclaimer.
		* 	Redistributions in binary form must reproduce the above copyright notice, this list of 
			conditions and the following disclaimer in the documentation.
		* 	Neither the name of the author nor the names of contributors may be used to endorse or 
			promote products derived from this software without specific prior written permission.
		* 	The use of any third party libraries, modules or data files provided with the source 
			form of the program must follow and respect the individual licensing terms of these 
			components. You do not have permission to modify or circumvent the licensing terms for 
			provided third party components unless the third party licenses specifically allow this. 
			If you do not agree to the terms of provided third party components then you do not have 
			permission to use them in your derived work(s).
		* 	The terms of this license do not apply to separate (but related) distributions if supplied 
			by the author(s), only to the contents of this source distribution, derived works, and the 
			binaries produced by either modified or unmodified versions of the source code. The terms 
			of this license do not also apply to any third party libraries or components provided with 
			the source distribution. Such third party elements are licensed under their own separate 
			agreements.

	THIS SOFTWARE IS PROVIDED BY DARRAGH COY ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
	BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL DARRAGH COY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
	IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

//=====================================================================================================================
/// @brief		Implementation of a vertex shader resource type.
/// @remarks	This is a template specialisation. Resource type implementions implement the resource type that 
///				is implicitly declared by the particular resource manager. For example a PixelShaderResourceManager
///				implicitly declares a resource type that stores pixel shaders and an implementation of this resource 
///				type (loading and freeing) must then be provided in source files.
//=====================================================================================================================

#include "Precompiled.h"
#include "Resources\ResourceManager.h"
#include "Files\DiskFile.h"

using namespace Game;

//=====================================================================================================================
/// @brief		Loads the pixel resource with the DirectX if possible.
/// @pre		The manager for this resource must have a valid Direct3D device
//=====================================================================================================================

template<> void Resource<IDirect3DVertexShader9*>::LoadData()
{
	// Make sure ok:
	
	ENSURE( m_manager_user_data != 0 , "The manager for this resource must have a valid Direct3D device" )
	
	// Free up everything first:
	
	FreeData();

	// See if we have a direct3d device
	
	IDirect3DDevice9 * device = (IDirect3DDevice9*) m_manager_user_data;
	
	// Open a disk file:
	
	DiskFile file; file.Open(m_name);
	
	// If we got data then load the shader:
	
	if ( file.IsOpen() && file.Size() > 0 )
	{
		// Load:
		
		HRESULT result = device->CreateVertexShader
		(
			(DWORD*) file.GetDataAhead() ,
			&m_data
		);		
		
		// See if success:
		
		if ( result == D3D_OK )
		{
			// Great: now try and get the constant table
			
			result = D3DXGetShaderConstantTable
			( 
				(DWORD*) file.GetDataAhead() , 
				(ID3DXConstantTable**) & m_user_data
			);
			
			// See if that succeded:
			
			if ( result == D3D_OK )
			{			
				// Good: ready to roll
			
				m_is_valid = true;
			}
		}	
		
		// If something went wrong then free everyting:
		
		if ( result != D3D_OK )
		{			
			// Free the constant table:
			
			if ( m_user_data != 0 )
			{
				((ID3DXConstantTable*)(m_user_data))->Release(); m_user_data = 0;
			}					
			
			// Free the shader itself if generated:
			
			if ( m_data != 0 )
			{
				m_data->Release(); m_data = 0;
			}				
		}
		
	}	// end if file opened ok
}

//=====================================================================================================================
/// @brief	Free's the texture resource with the DirectX manager if possible.
//=====================================================================================================================

template<> void Resource<IDirect3DVertexShader9*>::FreeData()
{	
	// See if there:
	
	if ( m_is_valid )
	{
		// Free the vertex shader:
		
		m_data->Release();
		
		// Free the constant table:
		
		((ID3DXConstantTable*)(m_user_data))->Release();
	}	
	
	// Clean up:
	
	m_data			= 0; 
	m_user_data		= 0;
	m_is_valid		= false;	
}

